# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re
from collections import OrderedDict
from contextlib import nullcontext

import numpy as np
import pytest

from astropy import table
from astropy import units as u
from astropy.coordinates import (
    BaseRepresentationOrDifferential,
    CartesianRepresentation,
    SkyCoord,
    SphericalRepresentation,
    StokesCoord,
    UnitSphericalRepresentation,
    search_around_3d,
)
from astropy.coordinates.earth import EarthLocation
from astropy.coordinates.tests.helper import skycoord_equal
from astropy.coordinates.tests.test_representation import representation_equal
from astropy.table import Column, MaskedColumn, QTable, Table, TableMergeError
from astropy.table.operations import _get_out_class, join_distance, join_skycoord
from astropy.time import Time, TimeDelta
from astropy.units.quantity import Quantity
from astropy.utils import metadata
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.metadata import MergeConflictError


def sort_eq(list1, list2):
    return sorted(list1) == sorted(list2)


def check_mask(col, exp_mask):
    """Check that col.mask == exp_mask"""
    if hasattr(col, "mask"):
        # Coerce expected mask into dtype of col.mask. In particular this is
        # needed for types like EarthLocation where the mask is a structured
        # array.
        exp_mask = np.array(exp_mask).astype(col.mask.dtype)
        out = np.all(col.mask == exp_mask)
    else:
        # With no mask the check is OK if all the expected mask values
        # are False (i.e. no auto-conversion to MaskedQuantity if it was
        # not required by the join).
        out = np.all(exp_mask == False)  # noqa: E712
    return out


class TestJoin:
    def _setup(self, t_cls=Table):
        lines1 = [
            " a   b   c ",
            "  0 foo  L1",
            "  1 foo  L2",
            "  1 bar  L3",
            "  2 bar  L4",
        ]
        lines2 = [
            " a   b   d ",
            "  1 foo  R1",
            "  1 foo  R2",
            "  2 bar  R3",
            "  4 bar  R4",
        ]
        self.t1 = t_cls.read(lines1, format="ascii")
        self.t2 = t_cls.read(lines2, format="ascii")
        self.t3 = t_cls(self.t2, copy=True)

        self.t1.meta.update(OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)]))
        self.t2.meta.update(OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)]))
        self.t3.meta.update(OrderedDict([("b", 3), ("c", [1, 2]), ("d", 2), ("a", 1)]))

        self.meta_merge = OrderedDict(
            [
                ("b", [1, 2, 3, 4]),
                ("c", {"a": 1, "b": 1}),
                ("d", 1),
                ("a", 1),
            ]
        )

    def test_table_meta_merge(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.join(self.t1, self.t2, join_type="inner")
        assert out.meta == self.meta_merge

    def test_table_meta_merge_conflict(self, operation_table_type):
        self._setup(operation_table_type)

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.join(self.t1, self.t3, join_type="inner")
        assert len(w) == 3

        assert out.meta == self.t3.meta

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.join(
                self.t1, self.t3, join_type="inner", metadata_conflicts="warn"
            )
        assert len(w) == 3

        assert out.meta == self.t3.meta

        out = table.join(
            self.t1, self.t3, join_type="inner", metadata_conflicts="silent"
        )

        assert out.meta == self.t3.meta

        with pytest.raises(MergeConflictError):
            out = table.join(
                self.t1, self.t3, join_type="inner", metadata_conflicts="error"
            )

        with pytest.raises(ValueError):
            out = table.join(
                self.t1, self.t3, join_type="inner", metadata_conflicts="nonsense"
            )

    def test_both_unmasked_inner(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2

        # Basic join with default parameters (inner join on common keys)
        t12 = table.join(t1, t2)
        assert type(t12) is operation_table_type
        assert type(t12["a"]) is type(t1["a"])
        assert type(t12["b"]) is type(t1["b"])
        assert type(t12["c"]) is type(t1["c"])
        assert type(t12["d"]) is type(t2["d"])
        assert t12.masked is False
        assert sort_eq(
            t12.pformat(),
            [
                " a   b   c   d ",
                "--- --- --- ---",
                "  1 foo  L2  R1",
                "  1 foo  L2  R2",
                "  2 bar  L4  R3",
            ],
        )
        # Table meta merged properly
        assert t12.meta == self.meta_merge

    def test_both_unmasked_left_right_outer(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2

        # Left join
        t12 = table.join(t1, t2, join_type="left")
        assert t12.has_masked_columns is True
        assert t12.masked is False
        for name in ("a", "b", "c"):
            assert type(t12[name]) is Column
        assert type(t12["d"]) is MaskedColumn
        assert sort_eq(
            t12.pformat(),
            [
                " a   b   c   d ",
                "--- --- --- ---",
                "  0 foo  L1  --",
                "  1 bar  L3  --",
                "  1 foo  L2  R1",
                "  1 foo  L2  R2",
                "  2 bar  L4  R3",
            ],
        )

        # Right join
        t12 = table.join(t1, t2, join_type="right")
        assert t12.has_masked_columns is True
        assert t12.masked is False
        assert sort_eq(
            t12.pformat(),
            [
                " a   b   c   d ",
                "--- --- --- ---",
                "  1 foo  L2  R1",
                "  1 foo  L2  R2",
                "  2 bar  L4  R3",
                "  4 bar  --  R4",
            ],
        )

        # Outer join
        t12 = table.join(t1, t2, join_type="outer")
        assert t12.has_masked_columns is True
        assert t12.masked is False
        assert sort_eq(
            t12.pformat(),
            [
                " a   b   c   d ",
                "--- --- --- ---",
                "  0 foo  L1  --",
                "  1 bar  L3  --",
                "  1 foo  L2  R1",
                "  1 foo  L2  R2",
                "  2 bar  L4  R3",
                "  4 bar  --  R4",
            ],
        )

        # Check that the common keys are 'a', 'b'
        t12a = table.join(t1, t2, join_type="outer")
        t12b = table.join(t1, t2, join_type="outer", keys=["a", "b"])
        assert np.all(t12a.as_array() == t12b.as_array())

    def test_both_unmasked_single_key_inner(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2

        # Inner join on 'a' column
        t12 = table.join(t1, t2, keys="a")
        assert type(t12) is operation_table_type
        assert type(t12["a"]) is type(t1["a"])
        assert type(t12["b_1"]) is type(t1["b"])
        assert type(t12["c"]) is type(t1["c"])
        assert type(t12["b_2"]) is type(t2["b"])
        assert type(t12["d"]) is type(t2["d"])
        assert t12.masked is False
        assert sort_eq(
            t12.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  1 foo  L2 foo  R1",
                "  1 foo  L2 foo  R2",
                "  1 bar  L3 foo  R1",
                "  1 bar  L3 foo  R2",
                "  2 bar  L4 bar  R3",
            ],
        )

    def test_both_unmasked_single_key_left_right_outer(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2

        # Left join
        t12 = table.join(t1, t2, join_type="left", keys="a")
        assert t12.has_masked_columns is True
        assert sort_eq(
            t12.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  0 foo  L1  --  --",
                "  1 foo  L2 foo  R1",
                "  1 foo  L2 foo  R2",
                "  1 bar  L3 foo  R1",
                "  1 bar  L3 foo  R2",
                "  2 bar  L4 bar  R3",
            ],
        )

        # Right join
        t12 = table.join(t1, t2, join_type="right", keys="a")
        assert t12.has_masked_columns is True
        assert sort_eq(
            t12.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  1 foo  L2 foo  R1",
                "  1 foo  L2 foo  R2",
                "  1 bar  L3 foo  R1",
                "  1 bar  L3 foo  R2",
                "  2 bar  L4 bar  R3",
                "  4  --  -- bar  R4",
            ],
        )

        # Outer join
        t12 = table.join(t1, t2, join_type="outer", keys="a")
        assert t12.has_masked_columns is True
        assert sort_eq(
            t12.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  0 foo  L1  --  --",
                "  1 foo  L2 foo  R1",
                "  1 foo  L2 foo  R2",
                "  1 bar  L3 foo  R1",
                "  1 bar  L3 foo  R2",
                "  2 bar  L4 bar  R3",
                "  4  --  -- bar  R4",
            ],
        )

    def test_masked_unmasked(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t1m = operation_table_type(self.t1, masked=True)
        t2 = self.t2

        # Result table is never masked
        t1m2 = table.join(t1m, t2, join_type="inner")
        assert t1m2.masked is False

        # Result should match non-masked result
        t12 = table.join(t1, t2)
        assert np.all(t12.as_array() == np.array(t1m2))

        # Mask out some values in left table and make sure they propagate
        t1m["b"].mask[1] = True
        t1m["c"].mask[2] = True
        t1m2 = table.join(t1m, t2, join_type="inner", keys="a")
        assert sort_eq(
            t1m2.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  1  --  L2 foo  R1",
                "  1  --  L2 foo  R2",
                "  1 bar  -- foo  R1",
                "  1 bar  -- foo  R2",
                "  2 bar  L4 bar  R3",
            ],
        )

        t21m = table.join(t2, t1m, join_type="inner", keys="a")
        assert sort_eq(
            t21m.pformat(),
            [
                " a  b_1  d  b_2  c ",
                "--- --- --- --- ---",
                "  1 foo  R2  --  L2",
                "  1 foo  R2 bar  --",
                "  1 foo  R1  --  L2",
                "  1 foo  R1 bar  --",
                "  2 bar  R3 bar  L4",
            ],
        )

    def test_masked_masked(self, operation_table_type):
        self._setup(operation_table_type)
        """Two masked tables"""
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        t1 = self.t1
        t1m = operation_table_type(self.t1, masked=True)
        t2 = self.t2
        t2m = operation_table_type(self.t2, masked=True)

        # Result table is never masked but original column types are preserved
        t1m2m = table.join(t1m, t2m, join_type="inner")
        assert t1m2m.masked is False
        for col in t1m2m.itercols():
            assert type(col) is MaskedColumn

        # Result should match non-masked result
        t12 = table.join(t1, t2)
        assert np.all(t12.as_array() == np.array(t1m2m))

        # Mask out some values in both tables and make sure they propagate
        t1m["b"].mask[1] = True
        t1m["c"].mask[2] = True
        t2m["d"].mask[2] = True
        t1m2m = table.join(t1m, t2m, join_type="inner", keys="a")
        assert sort_eq(
            t1m2m.pformat(),
            [
                " a  b_1  c  b_2  d ",
                "--- --- --- --- ---",
                "  1  --  L2 foo  R1",
                "  1  --  L2 foo  R2",
                "  1 bar  -- foo  R1",
                "  1 bar  -- foo  R2",
                "  2 bar  L4 bar  --",
            ],
        )

    def test_classes(self):
        """Ensure that classes and subclasses get through as expected"""

        class MyCol(Column):
            pass

        class MyMaskedCol(MaskedColumn):
            pass

        t1 = Table()
        t1["a"] = MyCol([1])
        t1["b"] = MyCol([2])
        t1["c"] = MyMaskedCol([3])

        t2 = Table()
        t2["a"] = Column([1, 2])
        t2["d"] = MyCol([3, 4])
        t2["e"] = MyMaskedCol([5, 6])

        t12 = table.join(t1, t2, join_type="inner")
        for name, exp_type in (
            ("a", MyCol),
            ("b", MyCol),
            ("c", MyMaskedCol),
            ("d", MyCol),
            ("e", MyMaskedCol),
        ):
            assert type(t12[name] is exp_type)

        t21 = table.join(t2, t1, join_type="left")
        # Note col 'b' gets upgraded from MyCol to MaskedColumn since it needs to be
        # masked, but col 'c' stays since MyMaskedCol supports masking.
        for name, exp_type in (
            ("a", MyCol),
            ("b", MaskedColumn),
            ("c", MyMaskedCol),
            ("d", MyCol),
            ("e", MyMaskedCol),
        ):
            assert type(t21[name] is exp_type)

    def test_col_rename(self, operation_table_type):
        self._setup(operation_table_type)
        """
        Test auto col renaming when there is a conflict.  Use
        non-default values of uniq_col_name and table_names.
        """
        t1 = self.t1
        t2 = self.t2
        t12 = table.join(
            t1,
            t2,
            uniq_col_name="x_{table_name}_{col_name}_y",
            table_names=["L", "R"],
            keys="a",
        )
        assert t12.colnames == ["a", "x_L_b_y", "c", "x_R_b_y", "d"]

    def test_rename_conflict(self, operation_table_type):
        self._setup(operation_table_type)
        """
        Test that auto-column rename fails because of a conflict
        with an existing column
        """
        t1 = self.t1
        t2 = self.t2
        t1["b_1"] = 1  # Add a new column b_1 that will conflict with auto-rename
        with pytest.raises(TableMergeError):
            table.join(t1, t2, keys="a")

    def test_missing_keys(self, operation_table_type):
        self._setup(operation_table_type)
        """Merge on a key column that doesn't exist"""
        t1 = self.t1
        t2 = self.t2
        with pytest.raises(TableMergeError):
            table.join(t1, t2, keys=["a", "not there"])

    def test_bad_join_type(self, operation_table_type):
        self._setup(operation_table_type)
        """Bad join_type input"""
        t1 = self.t1
        t2 = self.t2
        with pytest.raises(ValueError):
            table.join(t1, t2, join_type="illegal value")

    def test_no_common_keys(self, operation_table_type):
        self._setup(operation_table_type)
        """Merge tables with no common keys"""
        t1 = self.t1
        t2 = self.t2
        del t1["a"]
        del t1["b"]
        del t2["a"]
        del t2["b"]
        with pytest.raises(TableMergeError):
            table.join(t1, t2)

    def test_masked_key_column(self, operation_table_type):
        self._setup(operation_table_type)
        """Merge on a key column that has a masked element"""
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        t1 = self.t1
        t2 = operation_table_type(self.t2, masked=True)
        table.join(t1, t2)  # OK
        t2["a"].mask[0] = True
        with pytest.raises(TableMergeError):
            table.join(t1, t2)

    def test_col_meta_merge(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t2.rename_column("d", "c")  # force col conflict and renaming
        meta1 = OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)])
        meta2 = OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])

        # Key col 'a', should first value ('cm')
        t1["a"].unit = "cm"
        t2["a"].unit = "m"
        # Key col 'b', take first value 't1_b'
        t1["b"].info.description = "t1_b"
        # Key col 'b', take first non-empty value 't1_b'
        t2["b"].info.format = "%6s"
        # Key col 'a', should be merged meta
        t1["a"].info.meta = meta1
        t2["a"].info.meta = meta2
        # Key col 'b', should be meta2
        t2["b"].info.meta = meta2

        # All these should pass through
        t1["c"].info.format = "%3s"
        t1["c"].info.description = "t1_c"

        t2["c"].info.format = "%6s"
        t2["c"].info.description = "t2_c"

        if operation_table_type is Table:
            ctx = pytest.warns(
                metadata.MergeConflictWarning,
                match=(
                    r"In merged column 'a' the 'unit' attribute does not match \(cm"
                    r" != m\)"
                ),
            )
        else:
            ctx = nullcontext()

        with ctx:
            t12 = table.join(t1, t2, keys=["a", "b"])

        assert t12["a"].unit == "m"
        assert t12["b"].info.description == "t1_b"
        assert t12["b"].info.format == "%6s"
        assert t12["a"].info.meta == self.meta_merge
        assert t12["b"].info.meta == meta2
        assert t12["c_1"].info.format == "%3s"
        assert t12["c_1"].info.description == "t1_c"
        assert t12["c_2"].info.format == "%6s"
        assert t12["c_2"].info.description == "t2_c"

    def test_join_multidimensional(self, operation_table_type):
        self._setup(operation_table_type)

        # Regression test for #2984, which was an issue where join did not work
        # on multi-dimensional columns.

        t1 = operation_table_type()
        t1["a"] = [1, 2, 3]
        t1["b"] = np.ones((3, 4))

        t2 = operation_table_type()
        t2["a"] = [1, 2, 3]
        t2["c"] = [4, 5, 6]

        t3 = table.join(t1, t2)

        np.testing.assert_allclose(t3["a"], t1["a"])
        np.testing.assert_allclose(t3["b"], t1["b"])
        np.testing.assert_allclose(t3["c"], t2["c"])

    def test_join_multidimensional_masked(self, operation_table_type):
        self._setup(operation_table_type)
        """
        Test for outer join with multidimensional columns where masking is required.
        (Issue #4059).
        """
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")

        a = table.MaskedColumn([1, 2, 3], name="a")
        a2 = table.Column([1, 3, 4], name="a")
        b = table.MaskedColumn(
            [
                [1, 2],
                [3, 4],
                [5, 6],
            ],
            name="b",
            mask=[
                [1, 0],
                [0, 1],
                [0, 0],
            ],
        )
        c = table.Column(
            [
                [1, 1],
                [2, 2],
                [3, 3],
            ],
            name="c",
        )
        t1 = operation_table_type([a, b])
        t2 = operation_table_type([a2, c])
        t12 = table.join(t1, t2, join_type="inner")

        assert np.all(
            t12["b"].mask
            == [
                [True, False],
                [False, False],
            ]
        )
        assert not hasattr(t12["c"], "mask")

        t12 = table.join(t1, t2, join_type="outer")
        assert np.all(
            t12["b"].mask
            == [
                [True, False],
                [False, True],
                [False, False],
                [True, True],
            ]
        )
        assert np.all(
            t12["c"].mask
            == [
                [False, False],
                [True, True],
                [False, False],
                [False, False],
            ]
        )

    def test_mixin_functionality(self, mixin_cols):
        col = mixin_cols["m"]
        cls_name = type(col).__name__
        len_col = len(col)
        idx = np.arange(len_col)
        t1 = table.QTable([idx, col], names=["idx", "m1"])
        t2 = table.QTable([idx, col], names=["idx", "m2"])
        # Set up join mismatches for different join_type cases
        t1 = t1[[0, 1, 3]]
        t2 = t2[[0, 2, 3]]

        # Test inner join, which works for all mixin_cols
        out = table.join(t1, t2, join_type="inner")
        assert len(out) == 2
        assert out["m2"].__class__ is col.__class__
        assert np.all(out["idx"] == [0, 3])
        if cls_name == "SkyCoord":
            # SkyCoord doesn't support __eq__ so use our own
            assert skycoord_equal(out["m1"], col[[0, 3]])
            assert skycoord_equal(out["m2"], col[[0, 3]])
        elif "Repr" in cls_name or "Diff" in cls_name:
            assert np.all(representation_equal(out["m1"], col[[0, 3]]))
            assert np.all(representation_equal(out["m2"], col[[0, 3]]))
        else:
            assert np.all(out["m1"] == col[[0, 3]])
            assert np.all(out["m2"] == col[[0, 3]])

        # Check for left, right, outer join which requires masking. Works for
        # the listed mixins classes.
        if isinstance(col, (Quantity, Time, TimeDelta)):
            out = table.join(t1, t2, join_type="left")
            assert len(out) == 3
            assert np.all(out["idx"] == [0, 1, 3])
            assert np.all(out["m1"] == t1["m1"])
            assert np.all(out["m2"] == t2["m2"])
            check_mask(out["m1"], [False, False, False])
            check_mask(out["m2"], [False, True, False])

            out = table.join(t1, t2, join_type="right")
            assert len(out) == 3
            assert np.all(out["idx"] == [0, 2, 3])
            assert np.all(out["m1"] == t1["m1"])
            assert np.all(out["m2"] == t2["m2"])
            check_mask(out["m1"], [False, True, False])
            check_mask(out["m2"], [False, False, False])

            out = table.join(t1, t2, join_type="outer")
            assert len(out) == 4
            assert np.all(out["idx"] == [0, 1, 2, 3])
            assert np.all(out["m1"] == col)
            assert np.all(out["m2"] == col)
            assert check_mask(out["m1"], [False, False, True, False])
            assert check_mask(out["m2"], [False, True, False, False])
        else:
            # Otherwise make sure it fails with the right exception message
            for join_type in ("outer", "left", "right"):
                with pytest.raises(NotImplementedError) as err:
                    table.join(t1, t2, join_type=join_type)
                assert "join requires masking" in str(
                    err.value
                ) or "join unavailable" in str(err.value)

    def test_cartesian_join(self, operation_table_type):
        t1 = Table(rows=[(1, "a"), (2, "b")], names=["a", "b"])
        t2 = Table(rows=[(3, "c"), (4, "d")], names=["a", "c"])
        t12 = table.join(t1, t2, join_type="cartesian")

        assert t1.colnames == ["a", "b"]
        assert t2.colnames == ["a", "c"]
        assert len(t12) == len(t1) * len(t2)
        assert str(t12).splitlines() == [
            "a_1  b  a_2  c ",
            "--- --- --- ---",
            "  1   a   3   c",
            "  1   a   4   d",
            "  2   b   3   c",
            "  2   b   4   d",
        ]

        with pytest.raises(ValueError, match="cannot supply keys for a cartesian join"):
            t12 = table.join(t1, t2, join_type="cartesian", keys="a")

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    def test_join_with_join_skycoord_sky(self):
        sc1 = SkyCoord([0, 1, 1.1, 2], [0, 0, 0, 0], unit="deg")
        sc2 = SkyCoord([0.5, 1.05, 2.1], [0, 0, 0], unit="deg")
        t1 = Table([sc1], names=["sc"])
        t2 = Table([sc2], names=["sc"])
        t12 = table.join(t1, t2, join_funcs={"sc": join_skycoord(0.2 * u.deg)})
        exp = [
            "sc_id   sc_1    sc_2  ",
            "      deg,deg deg,deg ",
            "----- ------- --------",
            "    1 1.0,0.0 1.05,0.0",
            "    1 1.1,0.0 1.05,0.0",
            "    2 2.0,0.0  2.1,0.0",
        ]
        assert str(t12).splitlines() == exp

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    @pytest.mark.parametrize("distance_func", ["search_around_3d", search_around_3d])
    def test_join_with_join_skycoord_3d(self, distance_func):
        sc1 = SkyCoord([0, 1, 1.1, 2] * u.deg, [0, 0, 0, 0] * u.deg, [1, 1, 2, 1] * u.m)
        sc2 = SkyCoord([0.5, 1.05, 2.1] * u.deg, [0, 0, 0] * u.deg, [1, 1, 1] * u.m)
        t1 = Table([sc1], names=["sc"])
        t2 = Table([sc2], names=["sc"])
        join_func = join_skycoord(np.deg2rad(0.2) * u.m, distance_func=distance_func)
        t12 = table.join(t1, t2, join_funcs={"sc": join_func})
        exp = [
            "sc_id     sc_1        sc_2    ",
            "       deg,deg,m   deg,deg,m  ",
            "----- ----------- ------------",
            "    1 1.0,0.0,1.0 1.05,0.0,1.0",
            "    2 2.0,0.0,1.0  2.1,0.0,1.0",
        ]
        assert str(t12).splitlines() == exp

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    def test_join_with_join_distance_1d(self):
        c1 = [0, 1, 1.1, 2]
        c2 = [0.5, 1.05, 2.1]
        t1 = Table([c1], names=["col"])
        t2 = Table([c2], names=["col"])
        join_func = join_distance(
            0.2, kdtree_args={"leafsize": 32}, query_args={"p": 2}
        )
        t12 = table.join(t1, t2, join_type="outer", join_funcs={"col": join_func})
        exp = [
            "col_id col_1 col_2",
            "------ ----- -----",
            "     1   1.0  1.05",
            "     1   1.1  1.05",
            "     2   2.0   2.1",
            "     3   0.0    --",
            "     4    --   0.5",
        ]
        assert str(t12).splitlines() == exp

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    def test_join_with_join_distance_1d_multikey(self):
        from astropy.table.operations import _apply_join_funcs

        c1 = [0, 1, 1.1, 1.2, 2]
        id1 = [0, 1, 2, 2, 3]
        o1 = ["a", "b", "c", "d", "e"]
        c2 = [0.5, 1.05, 2.1]
        id2 = [0, 2, 4]
        o2 = ["z", "y", "x"]
        t1 = Table([c1, id1, o1], names=["col", "id", "o1"])
        t2 = Table([c2, id2, o2], names=["col", "id", "o2"])
        join_func = join_distance(0.2)
        join_funcs = {"col": join_func}
        t12 = table.join(t1, t2, join_type="outer", join_funcs=join_funcs)
        exp = [
            "col_id col_1  id  o1 col_2  o2",
            "------ ----- --- --- ----- ---",
            "     1   1.0   1   b    --  --",
            "     1   1.1   2   c  1.05   y",
            "     1   1.2   2   d  1.05   y",
            "     2   2.0   3   e    --  --",
            "     2    --   4  --   2.1   x",
            "     3   0.0   0   a    --  --",
            "     4    --   0  --   0.5   z",
        ]
        assert str(t12).splitlines() == exp

        left, right, keys = _apply_join_funcs(t1, t2, ("col", "id"), join_funcs)
        assert keys == ("col_id", "id")

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    def test_join_with_join_distance_1d_quantity(self):
        c1 = [0, 1, 1.1, 2] * u.m
        c2 = [500, 1050, 2100] * u.mm
        t1 = QTable([c1], names=["col"])
        t2 = QTable([c2], names=["col"])
        join_func = join_distance(20 * u.cm)
        t12 = table.join(t1, t2, join_funcs={"col": join_func})
        exp = [
            "col_id col_1 col_2 ",
            "         m     mm  ",
            "------ ----- ------",
            "     1   1.0 1050.0",
            "     1   1.1 1050.0",
            "     2   2.0 2100.0",
        ]
        assert str(t12).splitlines() == exp

        # Generate column name conflict
        t2["col_id"] = [0, 0, 0]
        t2["col__id"] = [0, 0, 0]
        t12 = table.join(t1, t2, join_funcs={"col": join_func})
        exp = [
            "col___id col_1 col_2  col_id col__id",
            "           m     mm                 ",
            "-------- ----- ------ ------ -------",
            "       1   1.0 1050.0      0       0",
            "       1   1.1 1050.0      0       0",
            "       2   2.0 2100.0      0       0",
        ]
        assert str(t12).splitlines() == exp

    @pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
    def test_join_with_join_distance_2d(self):
        c1 = np.array([[0, 1, 1.1, 2], [0, 0, 1, 0]]).transpose()
        c2 = np.array([[0.5, 1.05, 2.1], [0, 0, 0]]).transpose()
        t1 = Table([c1], names=["col"])
        t2 = Table([c2], names=["col"])
        join_func = join_distance(
            0.2, kdtree_args={"leafsize": 32}, query_args={"p": 2}
        )
        t12 = table.join(t1, t2, join_type="outer", join_funcs={"col": join_func})
        exp = [
            "col_id   col_1       col_2   ",
            f'{t12["col_id"].dtype.name}  float64[2]  float64[2]',  # int32 or int64
            "------ ---------- -----------",
            "     1 1.0 .. 0.0 1.05 .. 0.0",
            "     2 2.0 .. 0.0  2.1 .. 0.0",
            "     3 0.0 .. 0.0    -- .. --",
            "     4 1.1 .. 1.0    -- .. --",
            "     5   -- .. --  0.5 .. 0.0",
        ]
        assert t12.pformat(show_dtype=True) == exp

    def test_keys_left_right_basic(self):
        """Test using the keys_left and keys_right args to specify different
        join keys. This takes the standard test case but renames column 'a'
        to 'x' and 'y' respectively for tables 1 and 2. Then it compares the
        normal join on 'a' to the new join on 'x' and 'y'."""
        self._setup()

        for join_type in ("inner", "left", "right", "outer"):
            t1 = self.t1.copy()
            t2 = self.t2.copy()
            # Expected is same as joining on 'a' but with names 'x', 'y' instead
            t12_exp = table.join(t1, t2, keys="a", join_type=join_type)
            t12_exp.add_column(t12_exp["a"], name="x", index=1)
            t12_exp.add_column(t12_exp["a"], name="y", index=len(t1.colnames) + 1)
            del t12_exp["a"]

            # Different key names
            t1.rename_column("a", "x")
            t2.rename_column("a", "y")
            keys_left_list = ["x"]  # Test string key name
            keys_right_list = [["y"]]  # Test list of string key names
            if join_type == "outer":
                # Just do this for the outer join (others are the same)
                keys_left_list.append([t1["x"].tolist()])  # Test list key column
                keys_right_list.append([t2["y"]])  # Test Column key column

            for keys_left, keys_right in zip(keys_left_list, keys_right_list):
                t12 = table.join(
                    t1,
                    t2,
                    keys_left=keys_left,
                    keys_right=keys_right,
                    join_type=join_type,
                )

                assert t12.colnames == t12_exp.colnames
                for col in t12.values_equal(t12_exp).itercols():
                    assert np.all(col)
                assert t12_exp.meta == t12.meta

    def test_keys_left_right_exceptions(self):
        """Test exceptions using the keys_left and keys_right args to specify
        different join keys.
        """
        self._setup()
        t1 = self.t1
        t2 = self.t2

        msg = r"left table does not have key column 'z'"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left="z", keys_right=["a"])

        msg = r"left table has different length from key \[1, 2\]"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left=[[1, 2]], keys_right=["a"])

        msg = r"keys arg must be None if keys_left and keys_right are supplied"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left="z", keys_right=["a"], keys="a")

        msg = r"keys_left and keys_right args must have same length"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left=["a", "b"], keys_right=["a"])

        msg = r"keys_left and keys_right must both be provided"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left=["a", "b"])

        msg = r"cannot supply join_funcs arg and keys_left / keys_right"
        with pytest.raises(ValueError, match=msg):
            table.join(t1, t2, keys_left=["a"], keys_right=["a"], join_funcs={})

    def test_join_structured_column(self):
        """Regression tests for gh-13271."""
        # Two tables with matching names, including a structured column.
        t1 = Table(
            [
                np.array([(1.0, 1), (2.0, 2)], dtype=[("f", "f8"), ("i", "i8")]),
                ["one", "two"],
            ],
            names=["structured", "string"],
        )
        t2 = Table(
            [
                np.array([(2.0, 2), (4.0, 4)], dtype=[("f", "f8"), ("i", "i8")]),
                ["three", "four"],
            ],
            names=["structured", "string"],
        )
        t12 = table.join(t1, t2, ["structured"], join_type="outer")
        assert t12.pformat() == [
            "structured [f, i] string_1 string_2",
            "----------------- -------- --------",
            "          (1., 1)      one       --",
            "          (2., 2)      two    three",
            "          (4., 4)       --     four",
        ]


class TestSetdiff:
    def _setup(self, t_cls=Table):
        lines1 = [" a   b ", "  0 foo ", "  1 foo ", "  1 bar ", "  2 bar "]
        lines2 = [" a   b ", "  0 foo ", "  3 foo ", "  4 bar ", "  2 bar "]
        lines3 = [
            " a   b   d ",
            "  0 foo  R1",
            "  8 foo  R2",
            "  1 bar  R3",
            "  4 bar  R4",
        ]
        self.t1 = t_cls.read(lines1, format="ascii")
        self.t2 = t_cls.read(lines2, format="ascii")
        self.t3 = t_cls.read(lines3, format="ascii")

    def test_default_same_columns(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.setdiff(self.t1, self.t2)
        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert out.pformat() == [" a   b ", "--- ---", "  1 bar", "  1 foo"]

    def test_default_same_tables(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.setdiff(self.t1, self.t1)

        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert out.pformat() == [
            " a   b ",
            "--- ---",
        ]

    def test_extra_col_left_table(self, operation_table_type):
        self._setup(operation_table_type)

        with pytest.raises(ValueError):
            table.setdiff(self.t3, self.t1)

    def test_extra_col_right_table(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.setdiff(self.t1, self.t3)

        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert out.pformat() == [
            " a   b ",
            "--- ---",
            "  1 foo",
            "  2 bar",
        ]

    def test_keys(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.setdiff(self.t3, self.t1, keys=["a", "b"])

        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert out.pformat() == [
            " a   b   d ",
            "--- --- ---",
            "  4 bar  R4",
            "  8 foo  R2",
        ]

    def test_missing_key(self, operation_table_type):
        self._setup(operation_table_type)

        with pytest.raises(ValueError):
            table.setdiff(self.t3, self.t1, keys=["a", "d"])


class TestVStack:
    def _setup(self, t_cls=Table):
        self.t1 = t_cls.read(
            [
                " a   b",
                " 0. foo",
                " 1. bar",
            ],
            format="ascii",
        )

        self.t2 = t_cls.read(
            [
                " a    b   c",
                " 2.  pez  4",
                " 3.  sez  5",
            ],
            format="ascii",
        )

        self.t3 = t_cls.read(
            [
                " a    b",
                " 4.   7",
                " 5.   8",
                " 6.   9",
            ],
            format="ascii",
        )
        self.t4 = t_cls(self.t1, copy=True, masked=t_cls is Table)

        # The following table has meta-data that conflicts with t1
        self.t5 = t_cls(self.t1, copy=True)

        self.t1.meta.update(OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)]))
        self.t2.meta.update(OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)]))
        self.t4.meta.update(OrderedDict([("b", [5, 6]), ("c", {"c": 1}), ("e", 1)]))
        self.t5.meta.update(OrderedDict([("b", 3), ("c", "k"), ("d", 1)]))
        self.meta_merge = OrderedDict(
            [
                ("b", [1, 2, 3, 4, 5, 6]),
                ("c", {"a": 1, "b": 1, "c": 1}),
                ("d", 1),
                ("a", 1),
                ("e", 1),
            ]
        )

    def test_validate_join_type(self):
        self._setup()
        with pytest.raises(TypeError, match="Did you accidentally call vstack"):
            table.vstack(self.t1, self.t2)

    def test_stack_rows(self, operation_table_type):
        self._setup(operation_table_type)
        t2 = self.t1.copy()
        t2.meta.clear()
        out = table.vstack([self.t1, t2[1]])
        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert out.pformat() == [
            " a   b ",
            "--- ---",
            "0.0 foo",
            "1.0 bar",
            "1.0 bar",
        ]

    def test_stack_table_column(self, operation_table_type):
        self._setup(operation_table_type)
        t2 = self.t1.copy()
        t2.meta.clear()
        out = table.vstack([self.t1, t2["a"]])
        assert out.masked is False
        assert out.pformat() == [
            " a   b ",
            "--- ---",
            "0.0 foo",
            "1.0 bar",
            "0.0  --",
            "1.0  --",
        ]

    def test_table_meta_merge(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.vstack([self.t1, self.t2, self.t4], join_type="inner")
        assert out.meta == self.meta_merge

    def test_table_meta_merge_conflict(self, operation_table_type):
        self._setup(operation_table_type)

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.vstack([self.t1, self.t5], join_type="inner")
        assert len(w) == 2

        assert out.meta == self.t5.meta

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.vstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="warn"
            )
        assert len(w) == 2

        assert out.meta == self.t5.meta

        out = table.vstack(
            [self.t1, self.t5], join_type="inner", metadata_conflicts="silent"
        )

        assert out.meta == self.t5.meta

        with pytest.raises(MergeConflictError):
            out = table.vstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="error"
            )

        with pytest.raises(ValueError):
            out = table.vstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="nonsense"
            )

    def test_bad_input_type(self, operation_table_type):
        self._setup(operation_table_type)
        with pytest.raises(ValueError):
            table.vstack([])
        with pytest.raises(TypeError):
            table.vstack(1)
        with pytest.raises(TypeError):
            table.vstack([self.t2, 1])
        with pytest.raises(ValueError):
            table.vstack([self.t1, self.t2], join_type="invalid join type")

    def test_stack_basic_inner(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4

        t12 = table.vstack([t1, t2], join_type="inner")
        assert t12.masked is False
        assert type(t12) is operation_table_type
        assert type(t12["a"]) is type(t1["a"])
        assert type(t12["b"]) is type(t1["b"])
        assert t12.pformat() == [
            " a   b ",
            "--- ---",
            "0.0 foo",
            "1.0 bar",
            "2.0 pez",
            "3.0 sez",
        ]

        t124 = table.vstack([t1, t2, t4], join_type="inner")
        assert type(t124) is operation_table_type
        assert type(t12["a"]) is type(t1["a"])
        assert type(t12["b"]) is type(t1["b"])
        assert t124.pformat() == [
            " a   b ",
            "--- ---",
            "0.0 foo",
            "1.0 bar",
            "2.0 pez",
            "3.0 sez",
            "0.0 foo",
            "1.0 bar",
        ]

    def test_stack_basic_outer(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4
        t12 = table.vstack([t1, t2], join_type="outer")
        assert t12.masked is False
        assert t12.pformat() == [
            " a   b   c ",
            "--- --- ---",
            "0.0 foo  --",
            "1.0 bar  --",
            "2.0 pez   4",
            "3.0 sez   5",
        ]

        t124 = table.vstack([t1, t2, t4], join_type="outer")
        assert t124.masked is False
        assert t124.pformat() == [
            " a   b   c ",
            "--- --- ---",
            "0.0 foo  --",
            "1.0 bar  --",
            "2.0 pez   4",
            "3.0 sez   5",
            "0.0 foo  --",
            "1.0 bar  --",
        ]

    def test_stack_incompatible(self, operation_table_type):
        self._setup(operation_table_type)
        with pytest.raises(TableMergeError) as excinfo:
            table.vstack([self.t1, self.t3], join_type="inner")
        assert "The 'b' columns have incompatible types: {}".format(
            [self.t1["b"].dtype.name, self.t3["b"].dtype.name]
        ) in str(excinfo.value)

        with pytest.raises(TableMergeError) as excinfo:
            table.vstack([self.t1, self.t3], join_type="outer")
        assert "The 'b' columns have incompatible types:" in str(excinfo.value)

        with pytest.raises(TableMergeError):
            table.vstack([self.t1, self.t2], join_type="exact")

        t1_reshape = self.t1.copy()
        t1_reshape["b"].shape = [2, 1]
        with pytest.raises(TableMergeError) as excinfo:
            table.vstack([self.t1, t1_reshape])
        assert "have different shape" in str(excinfo.value)

    def test_vstack_one_masked(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t4 = self.t4
        t4["b"].mask[1] = True
        t14 = table.vstack([t1, t4])
        assert t14.masked is False
        assert t14.pformat() == [
            " a   b ",
            "--- ---",
            "0.0 foo",
            "1.0 bar",
            "0.0 foo",
            "1.0  --",
        ]

    def test_col_meta_merge_inner(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4

        # Key col 'a', should last value ('km')
        t1["a"].info.unit = "cm"
        t2["a"].info.unit = "m"
        t4["a"].info.unit = "km"

        # Key col 'a' format should take last when all match
        t1["a"].info.format = "%f"
        t2["a"].info.format = "%f"
        t4["a"].info.format = "%f"

        # Key col 'b', take first value 't1_b'
        t1["b"].info.description = "t1_b"

        # Key col 'b', take first non-empty value '%6s'
        t4["b"].info.format = "%6s"

        # Key col 'a', should be merged meta
        t1["a"].info.meta.update(
            OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)])
        )
        t2["a"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )
        t4["a"].info.meta.update(
            OrderedDict([("b", [5, 6]), ("c", {"c": 1}), ("e", 1)])
        )

        # Key col 'b', should be meta2
        t2["b"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )

        if operation_table_type is Table:
            ctx = pytest.warns(metadata.MergeConflictWarning)
        else:
            ctx = nullcontext()

        with ctx as warning_lines:
            out = table.vstack([t1, t2, t4], join_type="inner")

        if operation_table_type is Table:
            assert len(warning_lines) == 2
            assert (
                "In merged column 'a' the 'unit' attribute does not match (cm != m)"
                in str(warning_lines[0].message)
            )
            assert (
                "In merged column 'a' the 'unit' attribute does not match (m != km)"
                in str(warning_lines[1].message)
            )
            # Check units are suitably ignored for a regular Table
            assert out.pformat() == [
                "   a       b   ",
                "   km          ",
                "-------- ------",
                "0.000000    foo",
                "1.000000    bar",
                "2.000000    pez",
                "3.000000    sez",
                "0.000000    foo",
                "1.000000    bar",
            ]
        else:
            # Check QTable correctly dealt with units.
            assert out.pformat() == [
                "   a       b   ",
                "   km          ",
                "-------- ------",
                "0.000000    foo",
                "0.000010    bar",
                "0.002000    pez",
                "0.003000    sez",
                "0.000000    foo",
                "1.000000    bar",
            ]
        assert out["a"].info.unit == "km"
        assert out["a"].info.format == "%f"
        assert out["b"].info.description == "t1_b"
        assert out["b"].info.format == "%6s"
        assert out["a"].info.meta == self.meta_merge
        assert out["b"].info.meta == OrderedDict(
            [("b", [3, 4]), ("c", {"b": 1}), ("a", 1)]
        )

    def test_col_meta_merge_outer(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4

        # Key col 'a', should last value ('km')
        t1["a"].unit = "cm"
        t2["a"].unit = "m"
        t4["a"].unit = "km"

        # Key col 'a' format should take last when all match
        t1["a"].info.format = "%0d"
        t2["a"].info.format = "%0d"
        t4["a"].info.format = "%0d"

        # Key col 'b', take first value 't1_b'
        t1["b"].info.description = "t1_b"

        # Key col 'b', take first non-empty value '%6s'
        t4["b"].info.format = "%6s"

        # Key col 'a', should be merged meta
        t1["a"].info.meta.update(
            OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)])
        )
        t2["a"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )
        t4["a"].info.meta.update(
            OrderedDict([("b", [5, 6]), ("c", {"c": 1}), ("e", 1)])
        )

        # Key col 'b', should be meta2
        t2["b"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )

        # All these should pass through
        t2["c"].unit = "m"
        t2["c"].info.format = "%6s"
        t2["c"].info.description = "t2_c"

        with pytest.warns(metadata.MergeConflictWarning) as warning_lines:
            out = table.vstack([t1, t2, t4], join_type="outer")

        assert len(warning_lines) == 2
        assert (
            "In merged column 'a' the 'unit' attribute does not match (cm != m)"
            in str(warning_lines[0].message)
        )
        assert (
            "In merged column 'a' the 'unit' attribute does not match (m != km)"
            in str(warning_lines[1].message)
        )
        assert out["a"].unit == "km"
        assert out["a"].info.format == "%0d"
        assert out["b"].info.description == "t1_b"
        assert out["b"].info.format == "%6s"
        assert out["a"].info.meta == self.meta_merge
        assert out["b"].info.meta == OrderedDict(
            [("b", [3, 4]), ("c", {"b": 1}), ("a", 1)]
        )
        assert out["c"].info.unit == "m"
        assert out["c"].info.format == "%6s"
        assert out["c"].info.description == "t2_c"

    def test_vstack_one_table(self, operation_table_type):
        self._setup(operation_table_type)
        """Regression test for issue #3313"""
        assert (self.t1 == table.vstack(self.t1)).all()
        assert (self.t1 == table.vstack([self.t1])).all()

    def test_mixin_functionality(self, mixin_cols):
        col = mixin_cols["m"]
        len_col = len(col)
        t = table.QTable([col], names=["a"])
        cls_name = type(col).__name__

        # Vstack works for these classes:
        if isinstance(
            col,
            (
                u.Quantity,
                Time,
                TimeDelta,
                SkyCoord,
                EarthLocation,
                BaseRepresentationOrDifferential,
                StokesCoord,
            ),
        ):
            out = table.vstack([t, t])
            assert len(out) == len_col * 2
            if cls_name == "SkyCoord":
                # Argh, SkyCoord needs __eq__!!
                assert skycoord_equal(out["a"][len_col:], col)
                assert skycoord_equal(out["a"][:len_col], col)
            elif "Repr" in cls_name or "Diff" in cls_name:
                assert np.all(representation_equal(out["a"][:len_col], col))
                assert np.all(representation_equal(out["a"][len_col:], col))
            else:
                assert np.all(out["a"][:len_col] == col)
                assert np.all(out["a"][len_col:] == col)
        else:
            msg = f"vstack unavailable for mixin column type(s): {cls_name}"
            with pytest.raises(NotImplementedError, match=re.escape(msg)):
                table.vstack([t, t])

        # Check for outer stack which requires masking.  Only Time supports
        # this currently.
        t2 = table.QTable([col], names=["b"])  # different from col name for t
        if isinstance(col, (Time, TimeDelta, Quantity)):
            out = table.vstack([t, t2], join_type="outer")
            assert len(out) == len_col * 2
            assert np.all(out["a"][:len_col] == col)
            assert np.all(out["b"][len_col:] == col)
            assert check_mask(out["a"], [False] * len_col + [True] * len_col)
            assert check_mask(out["b"], [True] * len_col + [False] * len_col)
            # check directly stacking mixin columns:
            out2 = table.vstack([t, t2["b"]])
            assert np.all(out["a"] == out2["a"])
            assert np.all(out["b"] == out2["b"])
        else:
            with pytest.raises(NotImplementedError) as err:
                table.vstack([t, t2], join_type="outer")
            assert "vstack requires masking" in str(
                err.value
            ) or "vstack unavailable" in str(err.value)

    def test_vstack_different_representation(self):
        """Test that representations can be mixed together."""
        rep1 = CartesianRepresentation([1, 2] * u.km, [3, 4] * u.km, 1 * u.km)
        rep2 = SphericalRepresentation([0] * u.deg, [0] * u.deg, 10 * u.km)
        t1 = Table([rep1])
        t2 = Table([rep2])
        t12 = table.vstack([t1, t2])
        expected = CartesianRepresentation(
            [1, 2, 10] * u.km, [3, 4, 0] * u.km, [1, 1, 0] * u.km
        )
        assert np.all(representation_equal(t12["col0"], expected))

        rep3 = UnitSphericalRepresentation([0] * u.deg, [0] * u.deg)
        t3 = Table([rep3])
        with pytest.raises(ValueError, match="representations are inconsistent"):
            table.vstack([t1, t3])

    def test_vstack_structured_column(self):
        """Regression tests for gh-13271."""
        # Two tables with matching names, including a structured column.
        t1 = Table(
            [
                np.array([(1.0, 1), (2.0, 2)], dtype=[("f", "f8"), ("i", "i8")]),
                ["one", "two"],
            ],
            names=["structured", "string"],
        )
        t2 = Table(
            [
                np.array([(3.0, 3), (4.0, 4)], dtype=[("f", "f8"), ("i", "i8")]),
                ["three", "four"],
            ],
            names=["structured", "string"],
        )
        t12 = table.vstack([t1, t2])
        assert t12.pformat() == [
            "structured [f, i] string",
            "----------------- ------",
            "          (1., 1)    one",
            "          (2., 2)    two",
            "          (3., 3)  three",
            "          (4., 4)   four",
        ]

        # One table without the structured column.
        t3 = t2[("string",)]
        t13 = table.vstack([t1, t3])
        assert t13.pformat() == [
            "structured [f, i] string",
            "----------------- ------",
            "         (1.0, 1)    one",
            "         (2.0, 2)    two",
            "               --  three",
            "               --   four",
        ]

    def test_vstack_inputs_not_modified(self):
        """Tests that inputs are not modified, see issue #16119"""
        t1 = Table(data=dict(x=[1, 2, 3], y=["a", "b", "c"]))

        rows = list(t1)  # Table -> list of Rows
        rows0 = rows[0]
        t2 = table.vstack(rows)
        assert rows[0] is rows0

        tables = [t1, t1]
        t3 = table.vstack(tables)
        assert tables[0] is t1


class TestDStack:
    def _setup(self, t_cls=Table):
        self.t1 = t_cls.read(
            [
                " a   b",
                " 0. foo",
                " 1. bar",
            ],
            format="ascii",
        )

        self.t2 = t_cls.read(
            [
                " a    b   c",
                " 2.  pez  4",
                " 3.  sez  5",
            ],
            format="ascii",
        )
        self.t2["d"] = Time([1, 2], format="cxcsec")

        self.t3 = t_cls(
            {
                "a": [[5.0, 6.0], [4.0, 3.0]],
                "b": [["foo", "bar"], ["pez", "sez"]],
            },
            names=("a", "b"),
        )

        self.t4 = t_cls(self.t1, copy=True, masked=t_cls is Table)

        self.t5 = t_cls(
            {
                "a": [[4.0, 2.0], [1.0, 6.0]],
                "b": [["foo", "pez"], ["bar", "sez"]],
            },
            names=("a", "b"),
        )
        self.t6 = t_cls.read(
            [
                " a    b   c",
                " 7.  pez  2",
                " 4.  sez  6",
                " 6.  foo  3",
            ],
            format="ascii",
        )

    def test_validate_join_type(self):
        self._setup()
        with pytest.raises(TypeError, match="Did you accidentally call dstack"):
            table.dstack(self.t1, self.t2)

    @staticmethod
    def compare_dstack(tables, out):
        for ii, tbl in enumerate(tables):
            for name in out.columns:
                if name in tbl.colnames:
                    # Columns always compare equal
                    assert np.all(tbl[name] == out[name][:, ii])

                    # If input has a mask then output must have same mask
                    if hasattr(tbl[name], "mask"):
                        assert np.all(tbl[name].mask == out[name].mask[:, ii])

                    # If input has no mask then output might have a mask (if other table
                    # is missing that column). If so then all mask values should be False.
                    elif hasattr(out[name], "mask"):
                        assert not np.any(out[name].mask[:, ii])

                else:
                    # Column missing for this table, out must have a mask with all True.
                    assert np.all(out[name].mask[:, ii])

    def test_dstack_table_column(self, operation_table_type):
        """Stack a table with 3 cols and one column (gets auto-converted to Table)."""
        self._setup(operation_table_type)
        t2 = self.t1.copy()
        out = table.dstack([self.t1, t2["a"]])
        self.compare_dstack([self.t1, t2[("a",)]], out)

    def test_dstack_basic_outer(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail("Quantity columns do not support masking.")
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4
        t4["a"].mask[0] = True
        # Test for non-masked table
        t12 = table.dstack([t1, t2], join_type="outer")
        assert type(t12) is operation_table_type
        assert type(t12["a"]) is type(t1["a"])
        assert type(t12["b"]) is type(t1["b"])
        self.compare_dstack([t1, t2], t12)

        # Test for masked table
        t124 = table.dstack([t1, t2, t4], join_type="outer")
        assert type(t124) is operation_table_type
        assert type(t124["a"]) is type(t4["a"])
        assert type(t124["b"]) is type(t4["b"])
        self.compare_dstack([t1, t2, t4], t124)

    def test_dstack_basic_inner(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t4 = self.t4

        # Test for masked table
        t124 = table.dstack([t1, t2, t4], join_type="inner")
        assert type(t124) is operation_table_type
        assert type(t124["a"]) is type(t4["a"])
        assert type(t124["b"]) is type(t4["b"])
        self.compare_dstack([t1, t2, t4], t124)

    def test_dstack_multi_dimension_column(self, operation_table_type):
        self._setup(operation_table_type)
        t3 = self.t3
        t5 = self.t5
        t2 = self.t2
        t35 = table.dstack([t3, t5])
        assert type(t35) is operation_table_type
        assert type(t35["a"]) is type(t3["a"])
        assert type(t35["b"]) is type(t3["b"])
        self.compare_dstack([t3, t5], t35)

        with pytest.raises(TableMergeError):
            table.dstack([t2, t3])

    def test_dstack_different_length_table(self, operation_table_type):
        self._setup(operation_table_type)
        t2 = self.t2
        t6 = self.t6
        with pytest.raises(ValueError):
            table.dstack([t2, t6])

    def test_dstack_single_table(self):
        self._setup(Table)
        out = table.dstack(self.t1)
        assert np.all(out == self.t1)

    def test_dstack_representation(self):
        rep1 = SphericalRepresentation([1, 2] * u.deg, [3, 4] * u.deg, 1 * u.kpc)
        rep2 = SphericalRepresentation([10, 20] * u.deg, [30, 40] * u.deg, 10 * u.kpc)
        t1 = Table([rep1])
        t2 = Table([rep2])
        t12 = table.dstack([t1, t2])
        assert np.all(representation_equal(t12["col0"][:, 0], rep1))
        assert np.all(representation_equal(t12["col0"][:, 1], rep2))

    def test_dstack_skycoord(self):
        sc1 = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        sc2 = SkyCoord([10, 20] * u.deg, [30, 40] * u.deg)
        t1 = Table([sc1])
        t2 = Table([sc2])
        t12 = table.dstack([t1, t2])
        assert skycoord_equal(sc1, t12["col0"][:, 0])
        assert skycoord_equal(sc2, t12["col0"][:, 1])

    def test_dstack_structured_column(self):
        """Regression tests for gh-13271."""
        # Two tables with matching names, including a structured column.
        t1 = Table(
            [
                np.array([(1.0, 1), (2.0, 2)], dtype=[("f", "f8"), ("i", "i8")]),
                ["one", "two"],
            ],
            names=["structured", "string"],
        )
        t2 = Table(
            [
                np.array([(3.0, 3), (4.0, 4)], dtype=[("f", "f8"), ("i", "i8")]),
                ["three", "four"],
            ],
            names=["structured", "string"],
        )
        t12 = table.dstack([t1, t2])
        assert t12.pformat() == [
            "structured [f, i]     string   ",
            "------------------ ------------",
            "(1., 1) .. (3., 3) one .. three",
            "(2., 2) .. (4., 4)  two .. four",
        ]

        # One table without the structured column.
        t3 = t2[("string",)]
        t13 = table.dstack([t1, t3])
        assert t13.pformat() == [
            "structured [f, i]    string   ",
            "----------------- ------------",
            "   (1.0, 1) .. -- one .. three",
            "   (2.0, 2) .. --  two .. four",
        ]


class TestHStack:
    def _setup(self, t_cls=Table):
        self.t1 = t_cls.read(
            [
                " a    b",
                " 0. foo",
                " 1. bar",
            ],
            format="ascii",
        )

        self.t2 = t_cls.read(
            [
                " a    b   c",
                " 2.  pez  4",
                " 3.  sez  5",
            ],
            format="ascii",
        )

        self.t3 = t_cls.read(
            [
                " d    e",
                " 4.   7",
                " 5.   8",
                " 6.   9",
            ],
            format="ascii",
        )
        self.t4 = t_cls(self.t1, copy=True, masked=True)
        self.t4["a"].name = "f"
        self.t4["b"].name = "g"

        # The following table has meta-data that conflicts with t1
        self.t5 = t_cls(self.t1, copy=True)

        self.t1.meta.update(OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)]))
        self.t2.meta.update(OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)]))
        self.t4.meta.update(OrderedDict([("b", [5, 6]), ("c", {"c": 1}), ("e", 1)]))
        self.t5.meta.update(OrderedDict([("b", 3), ("c", "k"), ("d", 1)]))
        self.meta_merge = OrderedDict(
            [
                ("b", [1, 2, 3, 4, 5, 6]),
                ("c", {"a": 1, "b": 1, "c": 1}),
                ("d", 1),
                ("a", 1),
                ("e", 1),
            ]
        )

    def test_validate_join_type(self):
        self._setup()
        with pytest.raises(TypeError, match="Did you accidentally call hstack"):
            table.hstack(self.t1, self.t2)

    def test_stack_same_table(self, operation_table_type):
        """
        From #2995, test that hstack'ing references to the same table has the
        expected output.
        """
        self._setup(operation_table_type)
        out = table.hstack([self.t1, self.t1])
        assert out.masked is False
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2",
            "--- --- --- ---",
            "0.0 foo 0.0 foo",
            "1.0 bar 1.0 bar",
        ]

    def test_stack_rows(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.hstack([self.t1[0], self.t2[1]])
        assert out.masked is False
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2  c ",
            "--- --- --- --- ---",
            "0.0 foo 3.0 sez   5",
        ]

    def test_stack_columns(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.hstack([self.t1, self.t2["c"]])
        assert type(out["a"]) is type(self.t1["a"])
        assert type(out["b"]) is type(self.t1["b"])
        assert type(out["c"]) is type(self.t2["c"])
        assert out.pformat() == [
            " a   b   c ",
            "--- --- ---",
            "0.0 foo   4",
            "1.0 bar   5",
        ]

    def test_table_meta_merge(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.hstack([self.t1, self.t2, self.t4], join_type="inner")
        assert out.meta == self.meta_merge

    def test_table_meta_merge_conflict(self, operation_table_type):
        self._setup(operation_table_type)

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.hstack([self.t1, self.t5], join_type="inner")
        assert len(w) == 2

        assert out.meta == self.t5.meta

        with pytest.warns(metadata.MergeConflictWarning) as w:
            out = table.hstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="warn"
            )
        assert len(w) == 2

        assert out.meta == self.t5.meta

        out = table.hstack(
            [self.t1, self.t5], join_type="inner", metadata_conflicts="silent"
        )

        assert out.meta == self.t5.meta

        with pytest.raises(MergeConflictError):
            out = table.hstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="error"
            )

        with pytest.raises(ValueError):
            out = table.hstack(
                [self.t1, self.t5], join_type="inner", metadata_conflicts="nonsense"
            )

    def test_bad_input_type(self, operation_table_type):
        self._setup(operation_table_type)
        with pytest.raises(ValueError):
            table.hstack([])
        with pytest.raises(TypeError):
            table.hstack(1)
        with pytest.raises(TypeError):
            table.hstack([self.t2, 1])
        with pytest.raises(ValueError):
            table.hstack([self.t1, self.t2], join_type="invalid join type")

    def test_stack_basic(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = self.t2
        t3 = self.t3
        t4 = self.t4

        out = table.hstack([t1, t2], join_type="inner")
        assert out.masked is False
        assert type(out) is operation_table_type
        assert type(out["a_1"]) is type(t1["a"])
        assert type(out["b_1"]) is type(t1["b"])
        assert type(out["a_2"]) is type(t2["a"])
        assert type(out["b_2"]) is type(t2["b"])
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2  c ",
            "--- --- --- --- ---",
            "0.0 foo 2.0 pez   4",
            "1.0 bar 3.0 sez   5",
        ]

        # stacking as a list gives same result
        out_list = table.hstack([t1, t2], join_type="inner")
        assert out.pformat() == out_list.pformat()

        out = table.hstack([t1, t2], join_type="outer")
        assert out.pformat() == out_list.pformat()

        out = table.hstack([t1, t2, t3, t4], join_type="outer")
        assert out.masked is False
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2  c   d   e   f   g ",
            "--- --- --- --- --- --- --- --- ---",
            "0.0 foo 2.0 pez   4 4.0   7 0.0 foo",
            "1.0 bar 3.0 sez   5 5.0   8 1.0 bar",
            " --  --  --  --  -- 6.0   9  --  --",
        ]

        out = table.hstack([t1, t2, t3, t4], join_type="inner")
        assert out.masked is False
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2  c   d   e   f   g ",
            "--- --- --- --- --- --- --- --- ---",
            "0.0 foo 2.0 pez   4 4.0   7 0.0 foo",
            "1.0 bar 3.0 sez   5 5.0   8 1.0 bar",
        ]

    def test_stack_incompatible(self, operation_table_type):
        self._setup(operation_table_type)
        # For join_type exact, which will fail here because n_rows
        # does not match
        with pytest.raises(TableMergeError):
            table.hstack([self.t1, self.t3], join_type="exact")

    def test_hstack_one_masked(self, operation_table_type):
        if operation_table_type is QTable:
            pytest.xfail()
        self._setup(operation_table_type)
        t1 = self.t1
        t2 = operation_table_type(t1, copy=True, masked=True)
        t2.meta.clear()
        t2["b"].mask[1] = True
        out = table.hstack([t1, t2])
        assert out.pformat() == [
            "a_1 b_1 a_2 b_2",
            "--- --- --- ---",
            "0.0 foo 0.0 foo",
            "1.0 bar 1.0  --",
        ]

    def test_table_col_rename(self, operation_table_type):
        self._setup(operation_table_type)
        out = table.hstack(
            [self.t1, self.t2],
            join_type="inner",
            uniq_col_name="{table_name}_{col_name}",
            table_names=("left", "right"),
        )
        assert out.masked is False
        assert out.pformat() == [
            "left_a left_b right_a right_b  c ",
            "------ ------ ------- ------- ---",
            "   0.0    foo     2.0     pez   4",
            "   1.0    bar     3.0     sez   5",
        ]

    def test_col_meta_merge(self, operation_table_type):
        self._setup(operation_table_type)
        t1 = self.t1
        t3 = self.t3[:2]
        t4 = self.t4

        # Just set a bunch of meta and make sure it is the same in output
        meta1 = OrderedDict([("b", [1, 2]), ("c", {"a": 1}), ("d", 1)])
        t1["a"].unit = "cm"
        t1["b"].info.description = "t1_b"
        t4["f"].info.format = "%6s"
        t1["b"].info.meta.update(meta1)
        t3["d"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )
        t4["g"].info.meta.update(
            OrderedDict([("b", [5, 6]), ("c", {"c": 1}), ("e", 1)])
        )
        t3["e"].info.meta.update(
            OrderedDict([("b", [3, 4]), ("c", {"b": 1}), ("a", 1)])
        )
        t3["d"].unit = "m"
        t3["d"].info.format = "%6s"
        t3["d"].info.description = "t3_c"

        out = table.hstack([t1, t3, t4], join_type="exact")

        for t in [t1, t3, t4]:
            for name in t.colnames:
                for attr in ("meta", "unit", "format", "description"):
                    assert getattr(out[name].info, attr) == getattr(t[name].info, attr)

        # Make sure we got a copy of meta, not ref
        t1["b"].info.meta["b"] = None
        assert out["b"].info.meta["b"] == [1, 2]

    def test_hstack_one_table(self, operation_table_type):
        self._setup(operation_table_type)
        """Regression test for issue #3313"""
        assert (self.t1 == table.hstack(self.t1)).all()
        assert (self.t1 == table.hstack([self.t1])).all()

    def test_mixin_functionality(self, mixin_cols):
        col1 = mixin_cols["m"]
        col2 = col1[2:4]  # Shorter version of col1
        t1 = table.QTable([col1])
        t2 = table.QTable([col2])

        cls_name = type(col1).__name__

        out = table.hstack([t1, t2], join_type="inner")
        assert type(out["col0_1"]) is type(out["col0_2"])
        assert len(out) == len(col2)

        # Check that columns are as expected.
        if cls_name == "SkyCoord":
            assert skycoord_equal(out["col0_1"], col1[: len(col2)])
            assert skycoord_equal(out["col0_2"], col2)
        elif "Repr" in cls_name or "Diff" in cls_name:
            assert np.all(representation_equal(out["col0_1"], col1[: len(col2)]))
            assert np.all(representation_equal(out["col0_2"], col2))
        else:
            assert np.all(out["col0_1"] == col1[: len(col2)])
            assert np.all(out["col0_2"] == col2)

        # Time class supports masking, all other mixins do not
        if isinstance(col1, (Time, TimeDelta, Quantity)):
            out = table.hstack([t1, t2], join_type="outer")
            assert len(out) == len(t1)
            assert np.all(out["col0_1"] == col1)
            assert np.all(out["col0_2"][: len(col2)] == col2)
            assert check_mask(out["col0_2"], [False, False, True, True])

            # check directly stacking mixin columns:
            out2 = table.hstack([t1, t2["col0"]], join_type="outer")
            assert np.all(out["col0_1"] == out2["col0_1"])
            assert np.all(out["col0_2"] == out2["col0_2"])
        else:
            with pytest.raises(NotImplementedError) as err:
                table.hstack([t1, t2], join_type="outer")
            assert "hstack requires masking" in str(err.value)


def test_unique(operation_table_type):
    t = operation_table_type.read(
        [
            " a b  c  d",
            " 2 b 7.0 0",
            " 1 c 3.0 5",
            " 2 b 6.0 2",
            " 2 a 4.0 3",
            " 1 a 1.0 7",
            " 2 b 5.0 1",
            " 0 a 0.0 4",
            " 1 a 2.0 6",
            " 1 c 3.0 5",
        ],
        format="ascii",
    )

    tu = operation_table_type(np.sort(t[:-1]))

    t_all = table.unique(t)
    assert sort_eq(t_all.pformat(), tu.pformat())
    t_s = t.copy()
    del t_s["b", "c", "d"]
    t_all = table.unique(t_s)
    assert sort_eq(
        t_all.pformat(),
        [
            " a ",
            "---",
            "  0",
            "  1",
            "  2",
        ],
    )

    key1 = "a"
    t1a = table.unique(t, key1)
    assert sort_eq(
        t1a.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
            "  1   c 3.0   5",
            "  2   b 7.0   0",
        ],
    )
    t1b = table.unique(t, key1, keep="last")
    assert sort_eq(
        t1b.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
            "  1   c 3.0   5",
            "  2   b 5.0   1",
        ],
    )
    t1c = table.unique(t, key1, keep="none")
    assert sort_eq(
        t1c.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
        ],
    )

    key2 = ["a", "b"]
    t2a = table.unique(t, key2)
    assert sort_eq(
        t2a.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
            "  1   a 1.0   7",
            "  1   c 3.0   5",
            "  2   a 4.0   3",
            "  2   b 7.0   0",
        ],
    )

    t2b = table.unique(t, key2, keep="last")
    assert sort_eq(
        t2b.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
            "  1   a 2.0   6",
            "  1   c 3.0   5",
            "  2   a 4.0   3",
            "  2   b 5.0   1",
        ],
    )
    t2c = table.unique(t, key2, keep="none")
    assert sort_eq(
        t2c.pformat(),
        [
            " a   b   c   d ",
            "--- --- --- ---",
            "  0   a 0.0   4",
            "  2   a 4.0   3",
        ],
    )

    key2 = ["a", "a"]
    with pytest.raises(ValueError) as exc:
        t2a = table.unique(t, key2)
    assert exc.value.args[0] == "duplicate key names"

    with pytest.raises(ValueError) as exc:
        table.unique(t, key2, keep=True)
    assert exc.value.args[0] == "'keep' should be one of 'first', 'last', 'none'"

    t1_m = operation_table_type(t1a, masked=True)
    t1_m["a"].mask[1] = True

    with pytest.raises(ValueError) as exc:
        t1_mu = table.unique(t1_m)
    assert (
        exc.value.args[0] == "cannot use columns with masked values as keys; "
        "remove column 'a' from keys and rerun unique()"
    )

    t1_mu = table.unique(t1_m, silent=True)
    assert t1_mu.masked is False
    assert t1_mu.pformat() == [
        " a   b   c   d ",
        "--- --- --- ---",
        "  0   a 0.0   4",
        "  2   b 7.0   0",
        " --   c 3.0   5",
    ]

    with pytest.raises(ValueError):
        t1_mu = table.unique(t1_m, silent=True, keys="a")

    t1_m = operation_table_type(t, masked=True)
    t1_m["a"].mask[1] = True
    t1_m["d"].mask[3] = True

    # Test that multiple masked key columns get removed in the correct
    # order
    t1_mu = table.unique(t1_m, keys=["d", "a", "b"], silent=True)
    assert t1_mu.masked is False
    assert t1_mu.pformat() == [
        " a   b   c   d ",
        "--- --- --- ---",
        "  2   a 4.0  --",
        "  2   b 7.0   0",
        " --   c 3.0   5",
    ]


def test_vstack_bytes(operation_table_type):
    """
    Test for issue #5617 when vstack'ing bytes columns in Py3.
    This is really an upstream numpy issue numpy/numpy/#8403.
    """
    t = operation_table_type([[b"a"]], names=["a"])
    assert t["a"].itemsize == 1

    t2 = table.vstack([t, t])
    assert len(t2) == 2
    assert t2["a"].itemsize == 1


def test_vstack_unicode():
    """
    Test for problem related to issue #5617 when vstack'ing *unicode*
    columns.  In this case the character size gets multiplied by 4.
    """
    t = table.Table([["a"]], names=["a"])
    assert t["a"].itemsize == 4  # 4-byte / char for U dtype

    t2 = table.vstack([t, t])
    assert len(t2) == 2
    assert t2["a"].itemsize == 4


def test_join_mixins_time_quantity():
    """
    Test for table join using non-ndarray key columns.
    """
    tm1 = Time([2, 1, 2], format="cxcsec")
    q1 = [2, 1, 1] * u.m
    idx1 = [1, 2, 3]
    tm2 = Time([2, 3], format="cxcsec")
    q2 = [2, 3] * u.m
    idx2 = [10, 20]
    t1 = Table([tm1, q1, idx1], names=["tm", "q", "idx"])
    t2 = Table([tm2, q2, idx2], names=["tm", "q", "idx"])
    # Output:
    #
    # <Table length=4>
    #         tm            q    idx_1 idx_2
    #                       m
    #       object       float64 int64 int64
    # ------------------ ------- ----- -----
    # 0.9999999999969589     1.0     2    --
    #   2.00000000000351     1.0     3    --
    #   2.00000000000351     2.0     1    10
    #  3.000000000000469     3.0    --    20

    t12 = table.join(t1, t2, join_type="outer", keys=["tm", "q"])
    # Key cols are lexically sorted
    assert np.all(t12["tm"] == Time([1, 2, 2, 3], format="cxcsec"))
    assert np.all(t12["q"] == [1, 1, 2, 3] * u.m)
    assert np.all(t12["idx_1"] == np.ma.array([2, 3, 1, 0], mask=[0, 0, 0, 1]))
    assert np.all(t12["idx_2"] == np.ma.array([0, 0, 10, 20], mask=[1, 1, 0, 0]))


def test_join_mixins_not_sortable():
    """
    Test for table join using non-ndarray key columns that are not sortable.
    """
    sc = SkyCoord([1, 2], [3, 4], unit="deg,deg")
    t1 = Table([sc, [1, 2]], names=["sc", "idx1"])
    t2 = Table([sc, [10, 20]], names=["sc", "idx2"])

    with pytest.raises(TypeError, match="one or more key columns are not sortable"):
        table.join(t1, t2, keys="sc")


def test_join_non_1d_key_column():
    c1 = [[1, 2], [3, 4]]
    c2 = [1, 2]
    t1 = Table([c1, c2], names=["a", "b"])
    t2 = t1.copy()
    with pytest.raises(ValueError, match="key column 'a' must be 1-d"):
        table.join(t1, t2, keys="a")


def test_argsort_time_column():
    """Regression test for #10823."""
    times = Time(["2016-01-01", "2018-01-01", "2017-01-01"])
    t = Table([times], names=["time"])
    i = t.argsort("time")
    assert np.all(i == times.argsort())


def test_sort_indexed_table():
    """Test fix for #9473 and #6545 - and another regression test for #10823."""
    t = Table([[1, 3, 2], [6, 4, 5]], names=("a", "b"))
    t.add_index("a")
    t.sort("a")
    assert np.all(t["a"] == [1, 2, 3])
    assert np.all(t["b"] == [6, 5, 4])
    t.sort("b")
    assert np.all(t["b"] == [4, 5, 6])
    assert np.all(t["a"] == [3, 2, 1])

    times = ["2016-01-01", "2018-01-01", "2017-01-01"]
    tm = Time(times)
    t2 = Table([tm, [3, 2, 1]], names=["time", "flux"])
    t2.sort("flux")
    assert np.all(t2["flux"] == [1, 2, 3])
    t2.sort("time")
    assert np.all(t2["flux"] == [3, 1, 2])
    assert np.all(t2["time"] == tm[[0, 2, 1]])

    # Using the table as a TimeSeries implicitly sets the index, so
    # this test is a bit different from the above.
    from astropy.timeseries import TimeSeries

    ts = TimeSeries(time=times)
    ts["flux"] = [3, 2, 1]
    ts.sort("flux")
    assert np.all(ts["flux"] == [1, 2, 3])
    ts.sort("time")
    assert np.all(ts["flux"] == [3, 1, 2])
    assert np.all(ts["time"] == tm[[0, 2, 1]])


def test_get_out_class():
    c = table.Column([1, 2])
    mc = table.MaskedColumn([1, 2])
    q = [1, 2] * u.m

    assert _get_out_class([c, mc]) is mc.__class__
    assert _get_out_class([mc, c]) is mc.__class__
    assert _get_out_class([c, c]) is c.__class__
    assert _get_out_class([c]) is c.__class__

    with pytest.raises(ValueError):
        _get_out_class([c, q])

    with pytest.raises(ValueError):
        _get_out_class([q, c])


def test_masking_required_exception():
    """
    Test that outer join, hstack and vstack fail for a mixin column which
    does not support masking.
    """
    col = table.NdarrayMixin([0, 1, 2, 3])
    t1 = table.QTable([[1, 2, 3, 4], col], names=["a", "b"])
    t2 = table.QTable([[1, 2], col[:2]], names=["a", "c"])

    with pytest.raises(NotImplementedError) as err:
        table.vstack([t1, t2], join_type="outer")
    assert "vstack unavailable" in str(err.value)

    with pytest.raises(NotImplementedError) as err:
        table.hstack([t1, t2], join_type="outer")
    assert "hstack requires masking" in str(err.value)

    with pytest.raises(NotImplementedError) as err:
        table.join(t1, t2, join_type="outer")
    assert "join requires masking" in str(err.value)


def test_stack_columns():
    c = table.Column([1, 2])
    mc = table.MaskedColumn([1, 2])
    q = [1, 2] * u.m
    time = Time(["2001-01-02T12:34:56", "2001-02-03T00:01:02"])
    sc = SkyCoord([1, 2], [3, 4], unit="deg")
    cq = table.Column([11, 22], unit=u.m)

    t = table.hstack([c, q])
    assert t.__class__ is table.QTable
    assert t.masked is False
    t = table.hstack([q, c])
    assert t.__class__ is table.QTable
    assert t.masked is False

    t = table.hstack([mc, q])
    assert t.__class__ is table.QTable
    assert t.masked is False

    t = table.hstack([c, mc])
    assert t.__class__ is table.Table
    assert t.masked is False

    t = table.vstack([q, q])
    assert t.__class__ is table.QTable

    t = table.vstack([c, c])
    assert t.__class__ is table.Table

    t = table.hstack([c, time])
    assert t.__class__ is table.Table
    t = table.hstack([c, sc])
    assert t.__class__ is table.Table
    t = table.hstack([q, time, sc])
    assert t.__class__ is table.QTable

    with pytest.raises(ValueError):
        table.vstack([c, q])

    with pytest.raises(ValueError):
        t = table.vstack([q, cq])


def test_mixin_join_regression():
    # This used to trigger a ValueError:
    # ValueError: NumPy boolean array indexing assignment cannot assign
    # 6 input values to the 4 output values where the mask is true

    t1 = QTable()
    t1["index"] = [1, 2, 3, 4, 5]
    t1["flux1"] = [2, 3, 2, 1, 1] * u.Jy
    t1["flux2"] = [2, 3, 2, 1, 1] * u.Jy

    t2 = QTable()
    t2["index"] = [3, 4, 5, 6]
    t2["flux1"] = [2, 1, 1, 3] * u.Jy
    t2["flux2"] = [2, 1, 1, 3] * u.Jy

    t12 = table.join(t1, t2, keys=("index", "flux1", "flux2"), join_type="outer")

    assert len(t12) == 6


@pytest.mark.parametrize(
    "t1, t2",
    [
        # different names
        (
            Table([np.array([1])], names=["a"]),
            Table([np.array([1])], names=["b"]),
        ),
        # different data (broadcastable)
        (
            Table([np.array([])], names=["a"]),
            Table([np.array([1])], names=["a"]),
        ),
        # different data (not broadcastable)
        (
            Table([np.array([1, 2])], names=["a"]),
            Table([np.array([1, 2, 3])], names=["a"]),
        ),
        # different names and data (broadcastable)
        (
            Table([np.array([])], names=["a"]),
            Table([np.array([1])], names=["b"]),
        ),
        # different names and data (not broadcastable)
        (
            Table([np.array([1, 2])], names=["a"]),
            Table([np.array([1, 2, 3])], names=["b"]),
        ),
        # different data and array type (broadcastable)
        (
            Table([np.array([])], names=["a"]),
            Table([np.ma.MaskedArray([1])], names=["a"]),
        ),
        # different data and array type (not broadcastable)
        (
            Table([np.array([1, 2])], names=["a"]),
            Table([np.ma.MaskedArray([1, 2, 3])], names=["a"]),
        ),
    ],
)
def test_table_comp(t1, t2):
    # see https://github.com/astropy/astropy/issues/13421
    try:
        np.result_type(t1.dtype, t2.dtype)
        np.broadcast_shapes((len(t1),), (len(t2),))
    except (TypeError, ValueError):
        # dtypes are not comparable or arrays can't be broadcasted:
        # a simple bool should be returned
        assert not t1 == t2
        assert not t2 == t1
        assert t1 != t2
        assert t2 != t1
    else:
        # otherwise, the general case is to return a 1D array with dtype=bool
        assert not any(t1 == t2)
        assert not any(t2 == t1)
        assert all(t1 != t2)
        assert all(t2 != t1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.flrw.__init__.py`."""

import pytest

##############################################################################
# TESTS
##############################################################################


def test_getattr_error_attr_not_found():
    """Test getattr raises error for DNE."""
    with pytest.raises(ImportError):
        from astropy.cosmology.flrw import this_is_not_a_variable  # noqa: F401



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Affine coordinate transformations."""

from abc import abstractmethod

import numpy as np

from astropy.coordinates.transformations.base import CoordinateTransform

__all__ = [
    "BaseAffineTransform",
    "AffineTransform",
    "StaticMatrixTransform",
    "DynamicMatrixTransform",
]


class BaseAffineTransform(CoordinateTransform):
    """Base class for common functionality between the ``AffineTransform``-type
    subclasses.

    This base class is needed because `~astropy.coordinates.AffineTransform`
    and the matrix transform classes share the ``__call__()`` method, but
    differ in how they generate the affine parameters.
    `~astropy.coordinates.StaticMatrixTransform` passes in a matrix stored as a
    class attribute, and both of the matrix transforms pass in ``None`` for the
    offset. Hence, user subclasses would likely want to subclass this (rather
    than `~astropy.coordinates.AffineTransform`) if they want to provide
    alternative transformations using this machinery.

    """

    def _apply_transform(self, fromcoord, matrix, offset):
        from astropy.coordinates.representation import (
            CartesianDifferential,
            RadialDifferential,
            SphericalCosLatDifferential,
            SphericalDifferential,
            UnitSphericalRepresentation,
        )

        data = fromcoord.data
        has_velocity = "s" in data.differentials

        # Bail out if no transform is actually requested
        if matrix is None and offset is None:
            return data

        # list of unit differentials
        _unit_diffs = (
            SphericalDifferential._unit_differential,
            SphericalCosLatDifferential._unit_differential,
        )
        unit_vel_diff = has_velocity and isinstance(
            data.differentials["s"], _unit_diffs
        )
        rad_vel_diff = has_velocity and isinstance(
            data.differentials["s"], RadialDifferential
        )

        # Some initial checking to short-circuit doing any re-representation if
        # we're going to fail anyways:
        if isinstance(data, UnitSphericalRepresentation) and offset is not None:
            raise TypeError(
                "Position information stored on coordinate frame "
                "is insufficient to do a full-space position "
                "transformation (representation class: {data.__class__})"
            )

        if (
            has_velocity
            and (unit_vel_diff or rad_vel_diff)
            and offset is not None
            and "s" in offset.differentials
        ):
            # Coordinate has a velocity, but it is not a full-space velocity
            # that we need to do a velocity offset
            raise TypeError(
                "Velocity information stored on coordinate frame is insufficient to do"
                " a full-space velocity transformation (differential class:"
                f" {data.differentials['s'].__class__})"
            )

        if len(data.differentials) > 1:
            # We should never get here because the frame initializer shouldn't
            # allow more differentials, but this just adds protection for
            # subclasses that somehow skip the checks
            raise ValueError(
                "Representation passed to AffineTransform contains multiple associated"
                " differentials. Only a single differential with velocity units is"
                f" presently supported (differentials: {data.differentials})."
            )

        # If the representation is a UnitSphericalRepresentation, and this is
        # just a MatrixTransform, we have to try to turn the differential into a
        # Unit version of the differential (if no radial velocity) or a
        # sphericaldifferential with zero proper motion (if only a radial
        # velocity) so that the matrix operation works
        if (
            has_velocity
            and isinstance(data, UnitSphericalRepresentation)
            and not (unit_vel_diff or rad_vel_diff)
        ):
            # retrieve just velocity differential
            unit_diff = data.differentials["s"].represent_as(
                data.differentials["s"]._unit_differential, data
            )
            data = data.with_differentials({"s": unit_diff})  # updates key

        # If it's a RadialDifferential, we flat-out ignore the differentials
        # This is because, by this point (past the validation above), we can
        # only possibly be doing a rotation-only transformation, and that
        # won't change the radial differential. We later add it back in
        elif rad_vel_diff:
            data = data.without_differentials()

        # Convert the representation and differentials to cartesian without
        # having them attached to a frame
        rep = data.to_cartesian()
        diffs = {
            k: diff.represent_as(CartesianDifferential, data)
            for k, diff in data.differentials.items()
        }
        rep = rep.with_differentials(diffs)

        # Only do transform if matrix is specified. This is for speed in
        # transformations that only specify an offset (e.g., LSR)
        if matrix is not None:
            # Note: this applies to both representation and differentials
            rep = rep.transform(matrix)

        # TODO: if we decide to allow arithmetic between representations that
        # contain differentials, this can be tidied up
        newrep = rep.without_differentials()
        if offset is not None:
            newrep += offset.without_differentials()

        # We need a velocity (time derivative) and, for now, are strict: the
        # representation can only contain a velocity differential and no others.
        if has_velocity and not rad_vel_diff:
            veldiff = rep.differentials["s"]  # already in Cartesian form

            if offset is not None and "s" in offset.differentials:
                veldiff += offset.differentials["s"]

            newrep = newrep.with_differentials({"s": veldiff})

        if isinstance(fromcoord.data, UnitSphericalRepresentation):
            # Special-case this because otherwise the return object will think
            # it has a valid distance with the default return (a
            # CartesianRepresentation instance)

            if has_velocity and not unit_vel_diff and not rad_vel_diff:
                # We have to first represent as the Unit types we converted to,
                # then put the d_distance information back in to the
                # differentials and re-represent as their original forms
                newdiff = newrep.differentials["s"]
                _unit_cls = fromcoord.data.differentials["s"]._unit_differential
                newdiff = newdiff.represent_as(_unit_cls, newrep)

                kwargs = {comp: getattr(newdiff, comp) for comp in newdiff.components}
                kwargs["d_distance"] = fromcoord.data.differentials["s"].d_distance
                diffs = {
                    "s": type(fromcoord.data.differentials["s"])(copy=False, **kwargs)
                }

            elif has_velocity and unit_vel_diff:
                newdiff = newrep.differentials["s"].represent_as(
                    fromcoord.data.differentials["s"].__class__, newrep
                )
                diffs = {"s": newdiff}

            else:
                diffs = newrep.differentials

            newrep = newrep.represent_as(type(fromcoord.data)).with_differentials(diffs)

        elif has_velocity and unit_vel_diff:
            # Here, we're in the case where the representation is not
            # UnitSpherical, but the differential *is* one of the UnitSpherical
            # types. We have to convert back to that differential class or the
            # resulting frame will think it has a valid radial_velocity. This
            # can probably be cleaned up: we currently have to go through the
            # dimensional version of the differential before representing as the
            # unit differential so that the units work out (the distance length
            # unit shouldn't appear in the resulting proper motions)

            diff_cls = fromcoord.data.differentials["s"].__class__
            newrep = newrep.represent_as(
                type(fromcoord.data), diff_cls._dimensional_differential
            ).represent_as(type(fromcoord.data), diff_cls)

        # We pulled the radial differential off of the representation
        # earlier, so now we need to put it back. But, in order to do that, we
        # have to turn the representation into a repr that is compatible with
        # having a RadialDifferential
        if has_velocity and rad_vel_diff:
            newrep = newrep.represent_as(fromcoord.data.__class__)
            newrep = newrep.with_differentials({"s": fromcoord.data.differentials["s"]})

        return newrep

    def __call__(self, fromcoord, toframe):
        params = self._affine_params(fromcoord, toframe)
        newrep = self._apply_transform(fromcoord, *params)
        return toframe.realize_frame(newrep)

    @abstractmethod
    def _affine_params(self, fromcoord, toframe):
        pass


class AffineTransform(BaseAffineTransform):
    """
    A coordinate transformation specified as a function that yields a 3 x 3
    cartesian transformation matrix and a tuple of displacement vectors.

    See `~astropy.coordinates.Galactocentric` for
    an example.

    Parameters
    ----------
    transform_func : callable
        A callable that has the signature ``transform_func(fromcoord, toframe)``
        and returns: a (3, 3) matrix that operates on ``fromcoord`` in a
        Cartesian representation, and a ``CartesianRepresentation`` with
        (optionally) an attached velocity ``CartesianDifferential`` to represent
        a translation and offset in velocity to apply after the matrix
        operation.
    fromsys : class
        The coordinate frame class to start from.
    tosys : class
        The coordinate frame class to transform into.
    priority : float or int
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `~astropy.coordinates.TransformGraph` or None
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    Raises
    ------
    TypeError
        If ``transform_func`` is not callable

    """

    def __init__(self, transform_func, fromsys, tosys, priority=1, register_graph=None):
        if not callable(transform_func):
            raise TypeError("transform_func is not callable")
        self.transform_func = transform_func

        super().__init__(
            fromsys, tosys, priority=priority, register_graph=register_graph
        )

    def _affine_params(self, fromcoord, toframe):
        return self.transform_func(fromcoord, toframe)


class StaticMatrixTransform(BaseAffineTransform):
    """
    A coordinate transformation defined as a 3 x 3 cartesian
    transformation matrix.

    This is distinct from DynamicMatrixTransform in that this kind of matrix is
    independent of frame attributes.  That is, it depends *only* on the class of
    the frame.

    Parameters
    ----------
    matrix : array-like or callable
        A 3 x 3 matrix for transforming 3-vectors. In most cases will
        be unitary (although this is not strictly required). If a callable,
        will be called *with no arguments* to get the matrix.
    fromsys : class
        The coordinate frame class to start from.
    tosys : class
        The coordinate frame class to transform into.
    priority : float or int
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `~astropy.coordinates.TransformGraph` or None
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    Raises
    ------
    ValueError
        If the matrix is not 3 x 3

    """

    def __init__(self, matrix, fromsys, tosys, priority=1, register_graph=None):
        if callable(matrix):
            matrix = matrix()
        self.matrix = np.array(matrix)

        if self.matrix.shape != (3, 3):
            raise ValueError("Provided matrix is not 3 x 3")

        super().__init__(
            fromsys, tosys, priority=priority, register_graph=register_graph
        )

    def _affine_params(self, fromcoord, toframe):
        return self.matrix, None


class DynamicMatrixTransform(BaseAffineTransform):
    """
    A coordinate transformation specified as a function that yields a
    3 x 3 cartesian transformation matrix.

    This is similar to, but distinct from StaticMatrixTransform, in that the
    matrix for this class might depend on frame attributes.

    Parameters
    ----------
    matrix_func : callable
        A callable that has the signature ``matrix_func(fromcoord, toframe)`` and
        returns a 3 x 3 matrix that converts ``fromcoord`` in a cartesian
        representation to the new coordinate system.
    fromsys : class
        The coordinate frame class to start from.
    tosys : class
        The coordinate frame class to transform into.
    priority : float or int
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `~astropy.coordinates.TransformGraph` or None
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.

    Raises
    ------
    TypeError
        If ``matrix_func`` is not callable

    """

    def __init__(self, matrix_func, fromsys, tosys, priority=1, register_graph=None):
        if not callable(matrix_func):
            raise TypeError("matrix_func is not callable")
        self.matrix_func = matrix_func

        super().__init__(
            fromsys, tosys, priority=priority, register_graph=register_graph
        )

    def _affine_params(self, fromcoord, toframe):
        return self.matrix_func(fromcoord, toframe), None



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy.utils.data_info import ParentDtypeInfo


class NdarrayMixinInfo(ParentDtypeInfo):
    _represent_as_dict_primary_data = "data"

    def _represent_as_dict(self):
        """Represent Column as a dict that can be serialized."""
        col = self._parent
        out = {"data": col.view(np.ndarray)}
        return out

    def _construct_from_dict(self, map):
        """Construct Column from ``map``."""
        data = map.pop("data")
        out = self._parent_cls(data, **map)
        return out


class NdarrayMixin(np.ndarray):
    """
    Mixin column class to allow storage of arbitrary numpy
    ndarrays within a Table.  This is a subclass of numpy.ndarray
    and has the same initialization options as ``np.array()``.
    """

    info = NdarrayMixinInfo()

    def __new__(cls, obj, *args, **kwargs):
        self = np.array(obj, *args, **kwargs).view(cls)
        if "info" in getattr(obj, "__dict__", ()):
            self.info = obj.info
        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return

        if callable(super().__array_finalize__):
            super().__array_finalize__(obj)

        # Self was created from template (e.g. obj[slice] or (obj * 2))
        # or viewcast e.g. obj.view(Column).  In either case we want to
        # init Column attributes for self from obj if possible.
        if "info" in getattr(obj, "__dict__", ()):
            self.info = obj.info

    def __reduce__(self):
        # patch to pickle NdArrayMixin objects (ndarray subclasses), see
        # http://www.mail-archive.com/numpy-discussion@scipy.org/msg02446.html

        object_state = list(super().__reduce__())
        object_state[2] = (object_state[2], self.__dict__)
        return tuple(object_state)

    def __setstate__(self, state):
        # patch to unpickle NdarrayMixin objects (ndarray subclasses), see
        # http://www.mail-archive.com/numpy-discussion@scipy.org/msg02446.html

        nd_state, own_state = state
        super().__setstate__(nd_state)
        self.__dict__.update(own_state)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# LOCAL
from astropy.io.votable import parse
from astropy.utils.data import get_pkg_data_filename


def test_resource_groups():
    # Read the VOTABLE
    votable = parse(get_pkg_data_filename("data/resource_groups.xml"))

    resource = votable.resources[0]
    groups = resource.groups
    params = resource.params

    # Test that params inside groups are not outside

    assert len(groups[0].entries) == 1
    assert groups[0].entries[0].name == "ID"

    assert len(params) == 2
    assert params[0].name == "standardID"
    assert params[1].name == "accessURL"



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import copy
import os
import queue
import select
import socket
import threading
import time
import uuid
import warnings
import xmlrpc.client as xmlrpc
from urllib.parse import urlunparse

from astropy import log

from .constants import SAMP_STATUS_OK, __profile_version__
from .errors import SAMPHubError, SAMPProxyError, SAMPWarning
from .lockfile_helpers import create_lock_file, read_lockfile
from .standard_profile import ThreadingXMLRPCServer
from .utils import ServerProxyPool, _HubAsClient, internet_on
from .web_profile import WebProfileXMLRPCServer, web_profile_text_dialog

__all__ = ["SAMPHubServer", "WebProfileDialog"]

__doctest_skip__ = [".", "SAMPHubServer.*"]


class SAMPHubServer:
    """
    SAMP Hub Server.

    Parameters
    ----------
    secret : str, optional
        The secret code to use for the SAMP lockfile. If none is is specified,
        the :func:`uuid.uuid1` function is used to generate one.

    addr : str, optional
        Listening address (or IP). This defaults to 127.0.0.1 if the internet
        is not reachable, otherwise it defaults to the host name.

    port : int, optional
        Listening XML-RPC server socket port. If left set to 0 (the default),
        the operating system will select a free port.

    lockfile : str, optional
        Custom lockfile name.

    timeout : int, optional
        Hub inactivity timeout. If ``timeout > 0`` then the Hub automatically
        stops after an inactivity period longer than ``timeout`` seconds. By
        default ``timeout`` is set to 0 (Hub never expires).

    client_timeout : int, optional
        Client inactivity timeout. If ``client_timeout > 0`` then the Hub
        automatically unregisters the clients which result inactive for a
        period longer than ``client_timeout`` seconds. By default
        ``client_timeout`` is set to 0 (clients never expire).

    mode : str, optional
        Defines the Hub running mode. If ``mode`` is ``'single'`` then the Hub
        runs using the standard ``.samp`` lock-file, having a single instance
        for user desktop session. Otherwise, if ``mode`` is ``'multiple'``,
        then the Hub runs using a non-standard lock-file, placed in
        ``.samp-1`` directory, of the form ``samp-hub-<UUID>``, where
        ``<UUID>`` is a unique UUID assigned to the hub.

    label : str, optional
        A string used to label the Hub with a human readable name. This string
        is written in the lock-file assigned to the ``hub.label`` token.

    web_profile : bool, optional
        Enables or disables the Web Profile support.

    web_profile_dialog : class, optional
        Allows a class instance to be specified using ``web_profile_dialog``
        to replace the terminal-based message with e.g. a GUI pop-up. Two
        `queue.Queue` instances will be added to the instance as attributes
        ``queue_request`` and ``queue_result``. When a request is received via
        the ``queue_request`` queue, the pop-up should be displayed, and a
        value of `True` or `False` should be added to ``queue_result``
        depending on whether the user accepted or refused the connection.

    web_port : int, optional
        The port to use for web SAMP. This should not be changed except for
        testing purposes, since web SAMP should always use port 21012.

    pool_size : int, optional
        The number of socket connections opened to communicate with the
        clients.
    """

    def __init__(
        self,
        secret=None,
        addr=None,
        port=0,
        lockfile=None,
        timeout=0,
        client_timeout=0,
        mode="single",
        label="",
        web_profile=True,
        web_profile_dialog=None,
        web_port=21012,
        pool_size=20,
    ):
        # Generate random ID for the hub
        self._id = str(uuid.uuid1())

        # General settings
        self._is_running = False
        self._customlockfilename = lockfile
        self._lockfile = None
        self._addr = addr
        self._port = port
        self._mode = mode
        self._label = label
        self._timeout = timeout
        self._client_timeout = client_timeout
        self._pool_size = pool_size

        # Web profile specific attributes
        self._web_profile = web_profile
        self._web_profile_dialog = web_profile_dialog
        self._web_port = web_port

        self._web_profile_server = None
        self._web_profile_callbacks = {}
        self._web_profile_requests_queue = None
        self._web_profile_requests_result = None
        self._web_profile_requests_semaphore = None

        self._host_name = "127.0.0.1"
        if internet_on():
            try:
                self._host_name = socket.getfqdn()
                socket.getaddrinfo(self._addr or self._host_name, self._port or 0)
            except OSError:
                self._host_name = "127.0.0.1"

        # Threading stuff
        self._thread_lock = threading.Lock()
        self._thread_run = None
        self._thread_hub_timeout = None
        self._thread_client_timeout = None

        self._launched_threads = []

        # Variables for timeout testing:
        self._last_activity_time = None
        self._client_activity_time = {}

        # Hub message id counter, used to create hub msg ids
        self._hub_msg_id_counter = 0

        # Hub secret code
        self._hub_secret_code_customized = secret
        self._hub_secret = self._create_secret_code()

        # Hub public id (as SAMP client)
        self._hub_public_id = ""

        # Client ids
        # {private_key: (public_id, timestamp)}
        self._private_keys = {}

        # Metadata per client
        # {private_key: metadata}
        self._metadata = {}

        # List of subscribed clients per MType
        # {mtype: private_key list}
        self._mtype2ids = {}

        # List of subscribed MTypes per client
        # {private_key: mtype list}
        self._id2mtypes = {}

        # List of XML-RPC addresses per client
        # {public_id: (XML-RPC address, ServerProxyPool instance)}
        self._xmlrpc_endpoints = {}

        # Synchronous message id heap
        self._sync_msg_ids_heap = {}

        # Public ids counter
        self._client_id_counter = -1

    @property
    def id(self):
        """
        The unique hub ID.
        """
        return self._id

    def _register_standard_api(self, server):
        # Standard Profile only operations
        server.register_function(self._ping, "samp.hub.ping")
        server.register_function(
            self._set_xmlrpc_callback, "samp.hub.setXmlrpcCallback"
        )

        # Standard API operations
        server.register_function(self._register, "samp.hub.register")
        server.register_function(self._unregister, "samp.hub.unregister")
        server.register_function(self._declare_metadata, "samp.hub.declareMetadata")
        server.register_function(self._get_metadata, "samp.hub.getMetadata")
        server.register_function(
            self._declare_subscriptions, "samp.hub.declareSubscriptions"
        )
        server.register_function(self._get_subscriptions, "samp.hub.getSubscriptions")
        server.register_function(
            self._get_registered_clients, "samp.hub.getRegisteredClients"
        )
        server.register_function(
            self._get_subscribed_clients, "samp.hub.getSubscribedClients"
        )
        server.register_function(self._notify, "samp.hub.notify")
        server.register_function(self._notify_all, "samp.hub.notifyAll")
        server.register_function(self._call, "samp.hub.call")
        server.register_function(self._call_all, "samp.hub.callAll")
        server.register_function(self._call_and_wait, "samp.hub.callAndWait")
        server.register_function(self._reply, "samp.hub.reply")

    def _register_web_profile_api(self, server):
        # Web Profile methods like Standard Profile
        server.register_function(self._ping, "samp.webhub.ping")
        server.register_function(self._unregister, "samp.webhub.unregister")
        server.register_function(self._declare_metadata, "samp.webhub.declareMetadata")
        server.register_function(self._get_metadata, "samp.webhub.getMetadata")
        server.register_function(
            self._declare_subscriptions, "samp.webhub.declareSubscriptions"
        )
        server.register_function(
            self._get_subscriptions, "samp.webhub.getSubscriptions"
        )
        server.register_function(
            self._get_registered_clients, "samp.webhub.getRegisteredClients"
        )
        server.register_function(
            self._get_subscribed_clients, "samp.webhub.getSubscribedClients"
        )
        server.register_function(self._notify, "samp.webhub.notify")
        server.register_function(self._notify_all, "samp.webhub.notifyAll")
        server.register_function(self._call, "samp.webhub.call")
        server.register_function(self._call_all, "samp.webhub.callAll")
        server.register_function(self._call_and_wait, "samp.webhub.callAndWait")
        server.register_function(self._reply, "samp.webhub.reply")

        # Methods particularly for Web Profile
        server.register_function(self._web_profile_register, "samp.webhub.register")
        server.register_function(
            self._web_profile_allowReverseCallbacks, "samp.webhub.allowReverseCallbacks"
        )
        server.register_function(
            self._web_profile_pullCallbacks, "samp.webhub.pullCallbacks"
        )

    def _start_standard_server(self):
        self._server = ThreadingXMLRPCServer(
            (self._addr or self._host_name, self._port or 0),
            log,
            logRequests=False,
            allow_none=True,
        )
        prot = "http"

        self._port = self._server.socket.getsockname()[1]
        addr = f"{self._addr or self._host_name}:{self._port}"
        self._url = urlunparse((prot, addr, "", "", "", ""))
        self._server.register_introspection_functions()
        self._register_standard_api(self._server)

    def _start_web_profile_server(self):
        self._web_profile_requests_queue = queue.Queue(1)
        self._web_profile_requests_result = queue.Queue(1)
        self._web_profile_requests_semaphore = queue.Queue(1)

        if self._web_profile_dialog is not None:
            # TODO: Some sort of duck-typing on the web_profile_dialog object
            self._web_profile_dialog.queue_request = self._web_profile_requests_queue
            self._web_profile_dialog.queue_result = self._web_profile_requests_result

        try:
            self._web_profile_server = WebProfileXMLRPCServer(
                ("localhost", self._web_port), log, logRequests=False, allow_none=True
            )
            self._web_port = self._web_profile_server.socket.getsockname()[1]
            self._web_profile_server.register_introspection_functions()
            self._register_web_profile_api(self._web_profile_server)
            log.info("Hub set to run with Web Profile support enabled.")
        except OSError:
            log.warning(
                f"Port {self._web_port} already in use. Impossible to run the "
                "Hub with Web Profile support.",
                SAMPWarning,
            )
            self._web_profile = False
            # Cleanup
            self._web_profile_requests_queue = None
            self._web_profile_requests_result = None
            self._web_profile_requests_semaphore = None

    def _launch_thread(self, group=None, target=None, name=None, args=None):
        # Remove inactive threads
        remove = []
        for t in self._launched_threads:
            if not t.is_alive():
                remove.append(t)
        for t in remove:
            self._launched_threads.remove(t)

        # Start new thread
        t = threading.Thread(group=group, target=target, name=name, args=args)
        t.start()

        # Add to list of launched threads
        self._launched_threads.append(t)

    def _join_launched_threads(self, timeout=None):
        for t in self._launched_threads:
            t.join(timeout=timeout)

    def _timeout_test_hub(self):
        if self._timeout == 0:
            return

        last = time.time()
        while self._is_running:
            time.sleep(0.05)  # keep this small to check _is_running often
            now = time.time()
            if now - last > 1.0:
                with self._thread_lock:
                    if self._last_activity_time is not None:
                        if now - self._last_activity_time >= self._timeout:
                            warnings.warn(
                                "Timeout expired, Hub is shutting down!", SAMPWarning
                            )
                            self.stop()
                            return
                last = now

    def _timeout_test_client(self):
        if self._client_timeout == 0:
            return

        last = time.time()
        while self._is_running:
            time.sleep(0.05)  # keep this small to check _is_running often
            now = time.time()
            if now - last > 1.0:
                for private_key in self._client_activity_time.keys():
                    if (
                        now - self._client_activity_time[private_key]
                        > self._client_timeout
                        and private_key != self._hub_private_key
                    ):
                        warnings.warn(
                            f"Client {private_key} timeout expired!", SAMPWarning
                        )
                        self._notify_disconnection(private_key)
                        self._unregister(private_key)
                last = now

    def _hub_as_client_request_handler(self, method, args):
        if method == "samp.client.receiveCall":
            return self._receive_call(*args)
        elif method == "samp.client.receiveNotification":
            return self._receive_notification(*args)
        elif method == "samp.client.receiveResponse":
            return self._receive_response(*args)
        elif method == "samp.app.ping":
            return self._ping(*args)

    def _setup_hub_as_client(self):
        hub_metadata = {
            "samp.name": "Astropy SAMP Hub",
            "samp.description.text": self._label,
            "author.name": "The Astropy Collaboration",
            "samp.documentation.url": "https://docs.astropy.org/en/stable/samp",
            "samp.icon.url": self._url + "/samp/icon",
        }

        result = self._register(self._hub_secret)
        self._hub_public_id = result["samp.self-id"]
        self._hub_private_key = result["samp.private-key"]
        self._set_xmlrpc_callback(self._hub_private_key, self._url)
        self._declare_metadata(self._hub_private_key, hub_metadata)
        self._declare_subscriptions(
            self._hub_private_key, {"samp.app.ping": {}, "x-samp.query.by-meta": {}}
        )

    def start(self, wait=False):
        """
        Start the current SAMP Hub instance and create the lock file. Hub
        start-up can be blocking or non blocking depending on the ``wait``
        parameter.

        Parameters
        ----------
        wait : bool
            If `True` then the Hub process is joined with the caller, blocking
            the code flow. Usually `True` option is used to run a stand-alone
            Hub in an executable script. If `False` (default), then the Hub
            process runs in a separated thread. `False` is usually used in a
            Python shell.
        """
        if self._is_running:
            raise SAMPHubError("Hub is already running")

        if self._lockfile is not None:
            raise SAMPHubError("Hub is not running but lockfile is set")

        if self._web_profile:
            self._start_web_profile_server()

        self._start_standard_server()

        self._lockfile = create_lock_file(
            lockfilename=self._customlockfilename,
            mode=self._mode,
            hub_id=self.id,
            hub_params=self.params,
        )

        self._update_last_activity_time()
        self._setup_hub_as_client()

        self._start_threads()

        log.info("Hub started")

        if wait and self._is_running:
            self._thread_run.join()
            self._thread_run = None

    @property
    def params(self):
        """
        The hub parameters (which are written to the logfile).
        """
        params = {}

        # Keys required by standard profile

        params["samp.secret"] = self._hub_secret
        params["samp.hub.xmlrpc.url"] = self._url
        params["samp.profile.version"] = __profile_version__

        # Custom keys

        params["hub.id"] = self.id
        params["hub.label"] = self._label or f"Hub {self.id}"

        return params

    def _start_threads(self):
        self._thread_run = threading.Thread(target=self._serve_forever)
        self._thread_run.daemon = True

        if self._timeout > 0:
            self._thread_hub_timeout = threading.Thread(
                target=self._timeout_test_hub, name="Hub timeout test"
            )
            self._thread_hub_timeout.daemon = True
        else:
            self._thread_hub_timeout = None

        if self._client_timeout > 0:
            self._thread_client_timeout = threading.Thread(
                target=self._timeout_test_client, name="Client timeout test"
            )
            self._thread_client_timeout.daemon = True
        else:
            self._thread_client_timeout = None

        self._is_running = True
        self._thread_run.start()

        if self._thread_hub_timeout is not None:
            self._thread_hub_timeout.start()
        if self._thread_client_timeout is not None:
            self._thread_client_timeout.start()

    def _create_secret_code(self):
        if self._hub_secret_code_customized is not None:
            return self._hub_secret_code_customized
        else:
            return str(uuid.uuid1())

    def stop(self):
        """
        Stop the current SAMP Hub instance and delete the lock file.
        """
        if not self._is_running:
            return

        log.info("Hub is stopping...")

        self._notify_shutdown()

        self._is_running = False

        if self._lockfile and os.path.isfile(self._lockfile):
            lockfiledict = read_lockfile(self._lockfile)
            if lockfiledict["samp.secret"] == self._hub_secret:
                os.remove(self._lockfile)
        self._lockfile = None

        # Reset variables
        # TODO: What happens if not all threads are stopped after timeout?
        self._join_all_threads(timeout=10.0)

        self._hub_msg_id_counter = 0
        self._hub_secret = self._create_secret_code()
        self._hub_public_id = ""
        self._metadata = {}
        self._private_keys = {}
        self._mtype2ids = {}
        self._id2mtypes = {}
        self._xmlrpc_endpoints = {}
        self._last_activity_time = None

        log.info("Hub stopped.")

    def _join_all_threads(self, timeout=None):
        # In some cases, ``stop`` may be called from some of the sub-threads,
        # so we just need to make sure that we don't try and shut down the
        # calling thread.
        current_thread = threading.current_thread()
        if self._thread_run is not current_thread:
            self._thread_run.join(timeout=timeout)
            if not self._thread_run.is_alive():
                self._thread_run = None
        if (
            self._thread_hub_timeout is not None
            and self._thread_hub_timeout is not current_thread
        ):
            self._thread_hub_timeout.join(timeout=timeout)
            if not self._thread_hub_timeout.is_alive():
                self._thread_hub_timeout = None
        if (
            self._thread_client_timeout is not None
            and self._thread_client_timeout is not current_thread
        ):
            self._thread_client_timeout.join(timeout=timeout)
            if not self._thread_client_timeout.is_alive():
                self._thread_client_timeout = None

        self._join_launched_threads(timeout=timeout)

    @property
    def is_running(self):
        """Return an information concerning the Hub running status.

        Returns
        -------
        running : bool
            Is the hub running?
        """
        return self._is_running

    def _serve_forever(self):
        while self._is_running:
            try:
                read_ready = select.select([self._server.socket], [], [], 0.01)[0]
            except OSError as exc:
                warnings.warn(
                    f"Call to select() in SAMPHubServer failed: {exc}", SAMPWarning
                )
            else:
                if read_ready:
                    self._server.handle_request()

            if self._web_profile:
                # We now check if there are any connection requests from the
                # web profile, and if so, we initialize the pop-up.
                if self._web_profile_dialog is None:
                    try:
                        request = self._web_profile_requests_queue.get_nowait()
                    except queue.Empty:
                        pass
                    else:
                        web_profile_text_dialog(
                            request, self._web_profile_requests_result
                        )

                # We now check for requests over the web profile socket, and we
                # also update the pop-up in case there are any changes.
                try:
                    read_ready = select.select(
                        [self._web_profile_server.socket], [], [], 0.01
                    )[0]
                except OSError as exc:
                    warnings.warn(
                        f"Call to select() in SAMPHubServer failed: {exc}", SAMPWarning
                    )
                else:
                    if read_ready:
                        self._web_profile_server.handle_request()

        self._server.server_close()
        if self._web_profile_server is not None:
            self._web_profile_server.server_close()

    def _notify_shutdown(self):
        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.event.shutdown")
        for mtype in msubs:
            if mtype in self._mtype2ids:
                for key in self._mtype2ids[mtype]:
                    self._notify_(
                        self._hub_private_key,
                        self._private_keys[key][0],
                        {"samp.mtype": "samp.hub.event.shutdown", "samp.params": {}},
                    )

    def _notify_register(self, private_key):
        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.event.register")
        for mtype in msubs:
            if mtype in self._mtype2ids:
                public_id = self._private_keys[private_key][0]
                for key in self._mtype2ids[mtype]:
                    # if key != private_key:
                    self._notify(
                        self._hub_private_key,
                        self._private_keys[key][0],
                        {
                            "samp.mtype": "samp.hub.event.register",
                            "samp.params": {"id": public_id},
                        },
                    )

    def _notify_unregister(self, private_key):
        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.event.unregister")
        for mtype in msubs:
            if mtype in self._mtype2ids:
                public_id = self._private_keys[private_key][0]
                for key in self._mtype2ids[mtype]:
                    if key != private_key:
                        self._notify(
                            self._hub_private_key,
                            self._private_keys[key][0],
                            {
                                "samp.mtype": "samp.hub.event.unregister",
                                "samp.params": {"id": public_id},
                            },
                        )

    def _notify_metadata(self, private_key):
        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.event.metadata")
        for mtype in msubs:
            if mtype in self._mtype2ids:
                public_id = self._private_keys[private_key][0]
                for key in self._mtype2ids[mtype]:
                    # if key != private_key:
                    self._notify(
                        self._hub_private_key,
                        self._private_keys[key][0],
                        {
                            "samp.mtype": "samp.hub.event.metadata",
                            "samp.params": {
                                "id": public_id,
                                "metadata": self._metadata[private_key],
                            },
                        },
                    )

    def _notify_subscriptions(self, private_key):
        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.event.subscriptions")
        for mtype in msubs:
            if mtype in self._mtype2ids:
                public_id = self._private_keys[private_key][0]
                for key in self._mtype2ids[mtype]:
                    self._notify(
                        self._hub_private_key,
                        self._private_keys[key][0],
                        {
                            "samp.mtype": "samp.hub.event.subscriptions",
                            "samp.params": {
                                "id": public_id,
                                "subscriptions": self._id2mtypes[private_key],
                            },
                        },
                    )

    def _notify_disconnection(self, private_key):
        def _xmlrpc_call_disconnect(endpoint, private_key, hub_public_id, message):
            endpoint.samp.client.receiveNotification(
                private_key, hub_public_id, message
            )

        msubs = SAMPHubServer.get_mtype_subtypes("samp.hub.disconnect")
        public_id = self._private_keys[private_key][0]
        endpoint = self._xmlrpc_endpoints[public_id][1]

        for mtype in msubs:
            if mtype in self._mtype2ids and private_key in self._mtype2ids[mtype]:
                log.debug(f"notify disconnection to {public_id}")
                self._launch_thread(
                    target=_xmlrpc_call_disconnect,
                    args=(
                        endpoint,
                        private_key,
                        self._hub_public_id,
                        {
                            "samp.mtype": "samp.hub.disconnect",
                            "samp.params": {"reason": "Timeout expired!"},
                        },
                    ),
                )

    def _ping(self):
        self._update_last_activity_time()
        log.debug("ping")
        return "1"

    def _query_by_metadata(self, key, value):
        public_id_list = []
        for private_id in self._metadata:
            if key in self._metadata[private_id]:
                if self._metadata[private_id][key] == value:
                    public_id_list.append(self._private_keys[private_id][0])

        return public_id_list

    def _set_xmlrpc_callback(self, private_key, xmlrpc_addr):
        self._update_last_activity_time(private_key)
        if private_key in self._private_keys:
            if private_key == self._hub_private_key:
                public_id = self._private_keys[private_key][0]
                self._xmlrpc_endpoints[public_id] = (
                    xmlrpc_addr,
                    _HubAsClient(self._hub_as_client_request_handler),
                )
                return ""

            # Dictionary stored with the public id

            log.debug(f"set_xmlrpc_callback: {private_key} {xmlrpc_addr}")

            server_proxy_pool = None

            server_proxy_pool = ServerProxyPool(
                self._pool_size, xmlrpc.ServerProxy, xmlrpc_addr, allow_none=1
            )

            public_id = self._private_keys[private_key][0]
            self._xmlrpc_endpoints[public_id] = (xmlrpc_addr, server_proxy_pool)
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

        return ""

    def _perform_standard_register(self):
        with self._thread_lock:
            private_key, public_id = self._get_new_ids()
        self._private_keys[private_key] = (public_id, time.time())
        self._update_last_activity_time(private_key)
        self._notify_register(private_key)
        log.debug(f"register: private-key = {private_key} and self-id = {public_id}")
        return {
            "samp.self-id": public_id,
            "samp.private-key": private_key,
            "samp.hub-id": self._hub_public_id,
        }

    def _register(self, secret):
        self._update_last_activity_time()
        if secret == self._hub_secret:
            return self._perform_standard_register()
        else:
            # return {"samp.self-id": "", "samp.private-key": "", "samp.hub-id": ""}
            raise SAMPProxyError(7, "Bad secret code")

    def _get_new_ids(self):
        private_key = str(uuid.uuid1())
        self._client_id_counter += 1
        public_id = "cli#hub"
        if self._client_id_counter > 0:
            public_id = f"cli#{self._client_id_counter}"

        return private_key, public_id

    def _unregister(self, private_key):
        self._update_last_activity_time()

        public_key = ""

        self._notify_unregister(private_key)

        with self._thread_lock:
            if private_key in self._private_keys:
                public_key = self._private_keys[private_key][0]
                del self._private_keys[private_key]
            else:
                return ""

            if private_key in self._metadata:
                del self._metadata[private_key]

            if private_key in self._id2mtypes:
                del self._id2mtypes[private_key]

            for mtype in self._mtype2ids.keys():
                if private_key in self._mtype2ids[mtype]:
                    self._mtype2ids[mtype].remove(private_key)

            if public_key in self._xmlrpc_endpoints:
                del self._xmlrpc_endpoints[public_key]

            if private_key in self._client_activity_time:
                del self._client_activity_time[private_key]

            if self._web_profile:
                if private_key in self._web_profile_callbacks:
                    del self._web_profile_callbacks[private_key]
                self._web_profile_server.remove_client(private_key)

        log.debug(f"unregister {public_key} ({private_key})")

        return ""

    def _declare_metadata(self, private_key, metadata):
        self._update_last_activity_time(private_key)
        if private_key in self._private_keys:
            log.debug(f"declare_metadata: private-key = {private_key} {metadata = !s}")
            self._metadata[private_key] = metadata
            self._notify_metadata(private_key)
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")
        return ""

    def _get_metadata(self, private_key, client_id):
        self._update_last_activity_time(private_key)
        if private_key in self._private_keys:
            client_private_key = self._public_id_to_private_key(client_id)
            log.debug(
                f"get_metadata: private-key = {private_key} client-id = {client_id}"
            )
            if client_private_key is not None:
                if client_private_key in self._metadata:
                    log.debug(f"--> metadata = {self._metadata[client_private_key]}")
                    return self._metadata[client_private_key]
                else:
                    return {}
            else:
                raise SAMPProxyError(6, "Invalid client ID")
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _declare_subscriptions(self, private_key, mtypes):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            log.debug(
                f"declare_subscriptions: private-key = {private_key} {mtypes = !s}"
            )

            # remove subscription to previous mtypes
            if private_key in self._id2mtypes:
                prev_mtypes = self._id2mtypes[private_key]

                for mtype in prev_mtypes:
                    try:
                        self._mtype2ids[mtype].remove(private_key)
                    except ValueError:  # private_key is not in list
                        pass

            self._id2mtypes[private_key] = copy.deepcopy(mtypes)

            # remove duplicated MType for wildcard overwriting
            original_mtypes = copy.deepcopy(mtypes)

            for mtype in original_mtypes:
                if mtype.endswith("*"):
                    for mtype2 in original_mtypes:
                        if mtype2.startswith(mtype[:-1]) and mtype2 != mtype:
                            if mtype2 in mtypes:
                                del mtypes[mtype2]

            log.debug(
                "declare_subscriptions: subscriptions accepted from "
                f"{private_key} => {str(mtypes)}"
            )

            for mtype in mtypes:
                if mtype in self._mtype2ids:
                    if private_key not in self._mtype2ids[mtype]:
                        self._mtype2ids[mtype].append(private_key)
                else:
                    self._mtype2ids[mtype] = [private_key]

            self._notify_subscriptions(private_key)

        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

        return ""

    def _get_subscriptions(self, private_key, client_id):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            client_private_key = self._public_id_to_private_key(client_id)
            if client_private_key is not None:
                if client_private_key in self._id2mtypes:
                    log.debug(
                        f"get_subscriptions: client-id = {client_id} "
                        f"mtypes = {self._id2mtypes[client_private_key]!s}"
                    )
                    return self._id2mtypes[client_private_key]
                else:
                    log.debug(
                        f"get_subscriptions: client-id = {client_id} mtypes = missing"
                    )
                    return {}
            else:
                raise SAMPProxyError(6, "Invalid client ID")
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _get_registered_clients(self, private_key):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            reg_clients = []
            for pkey in self._private_keys.keys():
                if pkey != private_key:
                    reg_clients.append(self._private_keys[pkey][0])
            log.debug(
                f"get_registered_clients: {private_key = !s} clients = {reg_clients}"
            )
            return reg_clients
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _get_subscribed_clients(self, private_key, mtype):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            sub_clients = {}

            for pkey in self._private_keys.keys():
                if pkey != private_key and self._is_subscribed(pkey, mtype):
                    sub_clients[self._private_keys[pkey][0]] = {}

            log.debug(
                f"get_subscribed_clients: private_key = {private_key} mtype = {mtype} "
                f"clients = {sub_clients}"
            )
            return sub_clients
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    @staticmethod
    def get_mtype_subtypes(mtype):
        """
        Return a list containing all the possible wildcarded subtypes of MType.

        Parameters
        ----------
        mtype : str
            MType to be parsed.

        Returns
        -------
        types : list
            List of subtypes

        Examples
        --------
        >>> from astropy.samp import SAMPHubServer
        >>> SAMPHubServer.get_mtype_subtypes("samp.app.ping")
        ['samp.app.ping', 'samp.app.*', 'samp.*', '*']
        """
        subtypes = []

        msubs = mtype.split(".")
        indexes = list(range(len(msubs)))
        indexes.reverse()
        indexes.append(-1)

        for i in indexes:
            tmp_mtype = ".".join(msubs[: i + 1])
            if tmp_mtype != mtype:
                if tmp_mtype != "":
                    tmp_mtype = tmp_mtype + ".*"
                else:
                    tmp_mtype = "*"
            subtypes.append(tmp_mtype)

        return subtypes

    def _is_subscribed(self, private_key, mtype):
        subscribed = False

        msubs = SAMPHubServer.get_mtype_subtypes(mtype)

        for msub in msubs:
            if msub in self._mtype2ids:
                if private_key in self._mtype2ids[msub]:
                    subscribed = True

        return subscribed

    def _notify(self, private_key, recipient_id, message):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            if not (
                self._is_subscribed(
                    self._public_id_to_private_key(recipient_id), message["samp.mtype"]
                )
            ):
                raise SAMPProxyError(
                    2,
                    "Client {} not subscribed to MType {}".format(
                        recipient_id, message["samp.mtype"]
                    ),
                )

            self._launch_thread(
                target=self._notify_, args=(private_key, recipient_id, message)
            )
            return {}
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _notify_(self, sender_private_key, recipient_public_id, message):
        if sender_private_key not in self._private_keys:
            return

        sender_public_id = self._private_keys[sender_private_key][0]

        try:
            log.debug(
                "notify {} from {} to {}".format(
                    message["samp.mtype"], sender_public_id, recipient_public_id
                )
            )

            recipient_private_key = self._public_id_to_private_key(recipient_public_id)
            arg_params = (sender_public_id, message)
            samp_method_name = "receiveNotification"

            self._retry_method(
                recipient_private_key, recipient_public_id, samp_method_name, arg_params
            )

        except Exception as exc:
            warnings.warn(
                "{} notification from client {} to client {} failed [{}]".format(
                    message["samp.mtype"], sender_public_id, recipient_public_id, exc
                ),
                SAMPWarning,
            )

    def _notify_all(self, private_key, message):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            if "samp.mtype" not in message:
                raise SAMPProxyError(3, "samp.mtype keyword is missing")
            recipient_ids = self._notify_all_(private_key, message)
            return recipient_ids
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _notify_all_(self, sender_private_key, message):
        recipient_ids = []
        msubs = SAMPHubServer.get_mtype_subtypes(message["samp.mtype"])

        for mtype in msubs:
            if mtype in self._mtype2ids:
                for key in self._mtype2ids[mtype]:
                    if key != sender_private_key:
                        _recipient_id = self._private_keys[key][0]
                        recipient_ids.append(_recipient_id)
                        self._launch_thread(
                            target=self._notify,
                            args=(sender_private_key, _recipient_id, message),
                        )
        if not recipient_ids:
            warnings.warn(
                "No client was able to receive this message",
                SAMPWarning,
            )

        return recipient_ids

    def _call(self, private_key, recipient_id, msg_tag, message):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            if not (
                self._is_subscribed(
                    self._public_id_to_private_key(recipient_id), message["samp.mtype"]
                )
            ):
                raise SAMPProxyError(
                    2,
                    "Client {} not subscribed to MType {}".format(
                        recipient_id, message["samp.mtype"]
                    ),
                )
            public_id = self._private_keys[private_key][0]
            msg_id = self._get_new_hub_msg_id(public_id, msg_tag)
            self._launch_thread(
                target=self._call_,
                args=(private_key, public_id, recipient_id, msg_id, message),
            )
            return msg_id
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _call_(
        self, sender_private_key, sender_public_id, recipient_public_id, msg_id, message
    ):
        if sender_private_key not in self._private_keys:
            return

        try:
            log.debug(
                "call {} from {} to {} ({})".format(
                    msg_id.split(";;")[0],
                    sender_public_id,
                    recipient_public_id,
                    message["samp.mtype"],
                )
            )

            recipient_private_key = self._public_id_to_private_key(recipient_public_id)
            arg_params = (sender_public_id, msg_id, message)
            samp_methodName = "receiveCall"

            self._retry_method(
                recipient_private_key, recipient_public_id, samp_methodName, arg_params
            )

        except Exception as exc:
            warnings.warn(
                "{} call {} from client {} to client {} failed [{},{}]".format(
                    message["samp.mtype"],
                    msg_id.split(";;")[0],
                    sender_public_id,
                    recipient_public_id,
                    type(exc),
                    exc,
                ),
                SAMPWarning,
            )

    def _call_all(self, private_key, msg_tag, message):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            if "samp.mtype" not in message:
                raise SAMPProxyError(
                    3,
                    f"samp.mtype keyword is missing in message tagged as {msg_tag}",
                )

            public_id = self._private_keys[private_key][0]
            msg_id = self._call_all_(private_key, public_id, msg_tag, message)
            return msg_id
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _call_all_(self, sender_private_key, sender_public_id, msg_tag, message):
        msg_id = {}
        msubs = SAMPHubServer.get_mtype_subtypes(message["samp.mtype"])

        for mtype in msubs:
            if mtype in self._mtype2ids:
                for key in self._mtype2ids[mtype]:
                    if key != sender_private_key:
                        _msg_id = self._get_new_hub_msg_id(sender_public_id, msg_tag)
                        receiver_public_id = self._private_keys[key][0]
                        msg_id[receiver_public_id] = _msg_id
                        self._launch_thread(
                            target=self._call_,
                            args=(
                                sender_private_key,
                                sender_public_id,
                                receiver_public_id,
                                _msg_id,
                                message,
                            ),
                        )
        return msg_id

    def _call_and_wait(self, private_key, recipient_id, message, timeout):
        self._update_last_activity_time(private_key)

        if private_key in self._private_keys:
            timeout = int(timeout)

            now = time.time()
            response = {}

            msg_id = self._call(private_key, recipient_id, "samp::sync::call", message)
            self._sync_msg_ids_heap[msg_id] = None

            while self._is_running:
                if 0 < timeout <= time.time() - now:
                    del self._sync_msg_ids_heap[msg_id]
                    raise SAMPProxyError(1, "Timeout expired!")

                if self._sync_msg_ids_heap[msg_id] is not None:
                    response = copy.deepcopy(self._sync_msg_ids_heap[msg_id])
                    del self._sync_msg_ids_heap[msg_id]
                    break
                time.sleep(0.01)

            return response
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

    def _reply(self, private_key, msg_id, response):
        """
        The main method that gets called for replying. This starts up an
        asynchronous reply thread and returns.
        """
        self._update_last_activity_time(private_key)
        if private_key in self._private_keys:
            self._launch_thread(
                target=self._reply_, args=(private_key, msg_id, response)
            )
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")

        return {}

    def _reply_(self, responder_private_key, msg_id, response):
        if responder_private_key not in self._private_keys or not msg_id:
            return

        responder_public_id = self._private_keys[responder_private_key][0]
        counter, hub_public_id, recipient_public_id, recipient_msg_tag = msg_id.split(
            ";;", 3
        )

        try:
            log.debug(
                f"reply {counter} from {responder_public_id} to {recipient_public_id}"
            )

            if recipient_msg_tag == "samp::sync::call":
                if msg_id in self._sync_msg_ids_heap.keys():
                    self._sync_msg_ids_heap[msg_id] = response

            else:
                recipient_private_key = self._public_id_to_private_key(
                    recipient_public_id
                )
                arg_params = (responder_public_id, recipient_msg_tag, response)
                samp_method_name = "receiveResponse"

                self._retry_method(
                    recipient_private_key,
                    recipient_public_id,
                    samp_method_name,
                    arg_params,
                )

        except Exception as exc:
            warnings.warn(
                (
                    f"{recipient_msg_tag} reply from client {responder_public_id} "
                    f"to client {recipient_public_id} failed [{exc}]"
                ),
                SAMPWarning,
            )

    def _retry_method(
        self, recipient_private_key, recipient_public_id, samp_method_name, arg_params
    ):
        """
        This method is used to retry a SAMP call several times.

        Parameters
        ----------
        recipient_private_key
            The private key of the receiver of the call
        recipient_public_key
            The public key of the receiver of the call
        samp_method_name : str
            The name of the SAMP method to call
        arg_params : tuple
            Any additional arguments to be passed to the SAMP method
        """
        if recipient_private_key is None:
            raise SAMPHubError("Invalid client ID")

        from . import conf

        for attempt in range(conf.n_retries):
            if not self._is_running:
                time.sleep(0.01)
                continue

            try:
                if (
                    self._web_profile
                    and recipient_private_key in self._web_profile_callbacks
                ):
                    # Web Profile
                    callback = {
                        "samp.methodName": samp_method_name,
                        "samp.params": arg_params,
                    }
                    self._web_profile_callbacks[recipient_private_key].put(callback)

                else:
                    # Standard Profile
                    hub = self._xmlrpc_endpoints[recipient_public_id][1]
                    getattr(hub.samp.client, samp_method_name)(
                        recipient_private_key, *arg_params
                    )

            except xmlrpc.Fault as exc:
                log.debug(
                    f"{recipient_public_id} XML-RPC endpoint error "
                    f"(attempt {attempt + 1}): {exc.faultString}"
                )
                time.sleep(0.01)
            else:
                return

        # If we are here, then the above attempts failed
        error_message = (
            samp_method_name + " failed after " + str(conf.n_retries) + " attempts"
        )
        raise SAMPHubError(error_message)

    def _public_id_to_private_key(self, public_id):
        for private_key in self._private_keys.keys():
            if self._private_keys[private_key][0] == public_id:
                return private_key
        return None

    def _get_new_hub_msg_id(self, sender_public_id, sender_msg_id):
        with self._thread_lock:
            self._hub_msg_id_counter += 1
        return (
            f"msg#{self._hub_msg_id_counter};;{self._hub_public_id};;"
            f"{sender_public_id};;{sender_msg_id}"
        )

    def _update_last_activity_time(self, private_key=None):
        with self._thread_lock:
            self._last_activity_time = time.time()
            if private_key is not None:
                self._client_activity_time[private_key] = time.time()

    def _receive_notification(self, private_key, sender_id, message):
        return ""

    def _receive_call(self, private_key, sender_id, msg_id, message):
        if private_key == self._hub_private_key:
            if "samp.mtype" in message and message["samp.mtype"] == "samp.app.ping":
                self._reply(
                    self._hub_private_key,
                    msg_id,
                    {"samp.status": SAMP_STATUS_OK, "samp.result": {}},
                )

            elif "samp.mtype" in message and (
                message["samp.mtype"] == "x-samp.query.by-meta"
                or message["samp.mtype"] == "samp.query.by-meta"
            ):
                ids_list = self._query_by_metadata(
                    message["samp.params"]["key"], message["samp.params"]["value"]
                )
                self._reply(
                    self._hub_private_key,
                    msg_id,
                    {"samp.status": SAMP_STATUS_OK, "samp.result": {"ids": ids_list}},
                )

            return ""
        else:
            return ""

    def _receive_response(self, private_key, responder_id, msg_tag, response):
        return ""

    def _web_profile_register(
        self, identity_info, client_address=("unknown", 0), origin="unknown"
    ):
        self._update_last_activity_time()

        if client_address[0] not in ["localhost", "127.0.0.1"]:
            raise SAMPProxyError(403, "Request of registration rejected by the Hub.")

        if not origin:
            origin = "unknown"

        if isinstance(identity_info, dict):
            # an old version of the protocol provided just a string with the app name
            if "samp.name" not in identity_info:
                raise SAMPProxyError(
                    403,
                    "Request of registration rejected "
                    "by the Hub (application name not "
                    "provided).",
                )

        # Red semaphore for the other threads
        self._web_profile_requests_semaphore.put("wait")
        # Set the request to be displayed for the current thread
        self._web_profile_requests_queue.put((identity_info, client_address, origin))
        # Get the popup dialogue response
        response = self._web_profile_requests_result.get()
        # OK, semaphore green
        self._web_profile_requests_semaphore.get()

        if response:
            register_map = self._perform_standard_register()
            translator_url = "http://localhost:{}/translator/{}?ref=".format(
                self._web_port, register_map["samp.private-key"]
            )
            register_map["samp.url-translator"] = translator_url
            self._web_profile_server.add_client(register_map["samp.private-key"])
            return register_map
        else:
            raise SAMPProxyError(403, "Request of registration rejected by the user.")

    def _web_profile_allowReverseCallbacks(self, private_key, allow):
        self._update_last_activity_time()
        if private_key in self._private_keys:
            if allow == "0":
                if private_key in self._web_profile_callbacks:
                    del self._web_profile_callbacks[private_key]
            else:
                self._web_profile_callbacks[private_key] = queue.Queue()
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")
        return ""

    def _web_profile_pullCallbacks(self, private_key, timeout_secs):
        self._update_last_activity_time()
        if private_key in self._private_keys:
            callback = []
            callback_queue = self._web_profile_callbacks[private_key]
            try:
                while self._is_running:
                    item_queued = callback_queue.get_nowait()
                    callback.append(item_queued)
            except queue.Empty:
                pass
            return callback
        else:
            raise SAMPProxyError(5, f"Private-key {private_key} expired or invalid.")


class WebProfileDialog:
    """
    A base class to make writing Web Profile GUI consent dialogs
    easier.

    The concrete class must:

        1) Poll ``handle_queue`` periodically, using the timer services
           of the GUI's event loop.  This function will call
           ``self.show_dialog`` when a request requires authorization.
           ``self.show_dialog`` will be given the arguments:

              - ``samp_name``: The name of the application making the request.

              - ``details``: A dictionary of details about the client
                making the request.

              - ``client``: A hostname, port pair containing the client
                address.

              - ``origin``: A string containing the origin of the
                request.

        2) Call ``consent`` or ``reject`` based on the user's response to
           the dialog.
    """

    def handle_queue(self):
        try:
            request = self.queue_request.get_nowait()
        except queue.Empty:  # queue is set but empty
            pass
        except AttributeError:  # queue has not been set yet
            pass
        else:
            if isinstance(request[0], str):  # To support the old protocol version
                samp_name = request[0]
            else:
                samp_name = request[0]["samp.name"]

            self.show_dialog(samp_name, request[0], request[1], request[2])

    def consent(self):
        self.queue_result.put(True)

    def reject(self):
        self.queue_result.put(False)
